
#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

import click

from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL,\
    ChisubmitException
from chisubmit.common.utils import create_connection
from chisubmit.cli.shared.course import shared_course_list,\
    shared_course_set_user_attribute
import operator
from chisubmit.client.exceptions import UnknownObjectException
from chisubmit.cli.common import get_course_or_exit, get_user_or_exit,\
    api_obj_set_attribute, get_team_or_exit, catch_chisubmit_exceptions,\
    require_config, get_student_or_exit
import csv


@click.group(name="course")
@click.pass_context
def admin_course(ctx):
    pass


@click.command(name="add")
@click.argument('course_id', type=str)
@click.argument('name', type=str)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_add(ctx, course_id, name):
    try:
        course = ctx.obj["client"].get_course(course_id = course_id)
        print "ERROR: Cannot create course."
        print "       Course with course_id = %s already exists." % course_id
        ctx.exit(CHISUBMIT_FAIL)
    except UnknownObjectException, uoe:
        course = ctx.obj["client"].create_course(course_id = course_id,
                                                 name = name)
    
    return CHISUBMIT_SUCCESS


@click.command(name="remove")
@click.argument('course_id', type=str)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_remove(ctx, course_id):
    course = get_course_or_exit(ctx, course_id)
    
    course.delete()

    return CHISUBMIT_SUCCESS


@click.command(name="show")
@click.argument('course_id', type=str)
@click.option("--include-users", is_flag=True)
@click.option("--include-assignments", is_flag=True)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_show(ctx, course_id, include_users, include_assignments):
    course = get_course_or_exit(ctx, course_id) 
    
    print course.course_id, course.name

    if include_users:
        print
        print "INSTRUCTORS"
        print "-----------"
        for i in course.get_instructors():
            print "%s: %s, %s <%s>" % (i.user.username, i.user.last_name, i.user.first_name, i.user.email)
        print
            
        print "GRADERS"
        print "-------"
        for g in course.get_graders():
            print "%s: %s, %s <%s>" % (g.user.username, g.user.last_name, g.user.first_name, g.user.email)
        print
        
        print "STUDENTS"
        print "--------"
        for s in course.get_students():
            print "%s: %s, %s <%s>" % (s.user.username, s.user.last_name, s.user.first_name, s.user.email)
        print

    if include_assignments:
        print "ASSIGNMENTS"
        print "-----------"
        for a in course.get_assignments():
            print "%s: %s (Due: %s)" % (a.assignment_id, a.name, a.deadline)
        print

    
    return CHISUBMIT_SUCCESS




@click.command(name="add-instructor")
@click.argument('course_id', type=str)
@click.argument('user_id', type=str)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_add_instructor(ctx, course_id, user_id):
    course = get_course_or_exit(ctx, course_id)    
    user = get_user_or_exit(ctx, user_id)    
    course.add_instructor(user)
    
@click.command(name="add-grader")
@click.argument('course_id', type=str)
@click.argument('user_id', type=str)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_add_grader(ctx, course_id, user_id):
    course = get_course_or_exit(ctx, course_id)    
    user = get_user_or_exit(ctx, user_id)    
    course.add_grader(user)
    
@click.command(name="add-student")
@click.argument('course_id', type=str)
@click.argument('user_id', type=str)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_add_student(ctx, course_id, user_id):
    course = get_course_or_exit(ctx, course_id)    
    user = get_user_or_exit(ctx, user_id)    
    course.add_student(user)    
    
    
VALID_USER_TYPES = ['student', 'grader', 'instructor']    
    
@click.command(name="load-users")
@click.argument('course_id', type=str)
@click.argument('csv_file', type=click.File('rb'))
@click.argument('csv_username_column', type=str)
@click.argument('csv_fname_column', type=str)
@click.argument('csv_lname_column', type=str)
@click.argument('csv_email_column', type=str)
@click.option('--dry-run', is_flag=True)
@click.option('--sync', is_flag=True)
@click.option('--user-type', type=click.Choice(VALID_USER_TYPES + ['column']))
@click.option('--user-type-column', type=str)
@click.option('--id-from-email', is_flag=True)
@require_config
@click.pass_context
def admin_course_load_users(ctx, course_id, csv_file, csv_username_column, csv_fname_column, csv_lname_column, csv_email_column,
                                 dry_run, sync, user_type, user_type_column, id_from_email):   
    course = get_course_or_exit(ctx, course_id)    
                
    if user_type == "column" and user_type_column is None:
        print "You must specify a column with --user-type-column when using '--user-type column'"
        ctx.exit(CHISUBMIT_FAIL)
                
    csvf = csv.DictReader(csv_file)
            
    columns = [csv_username_column, csv_fname_column, csv_lname_column, csv_email_column]
    
    if user_type == "column":
        columns.append(user_type_column)
            
    for col in columns:
        if col not in csvf.fieldnames:
            print "CSV file %s does not have a '%s' column" % (csv_file, col)
            ctx.exit(CHISUBMIT_FAIL)
        
    student_usernames = set()
        
    for entry in csvf:
        username = entry[csv_username_column]
        
        if id_from_email:
            username = username.split("@")[0].strip()
        
        first_name = entry[csv_fname_column]
        last_name = entry[csv_lname_column]
        email = entry[csv_email_column]
        
        print "Processing %s (%s, %s)" % (username, last_name, first_name)        
        
        if user_type == "column":
            cur_user_type = entry[user_type_column]
            if cur_user_type not in VALID_USER_TYPES:
                print "- User %s has invalid user type '%s'." % (username, cur_user_type)
                continue
        else:
            cur_user_type = user_type

        if cur_user_type == "student":
            student_usernames.add(username)

        try:
            user = ctx.obj["client"].get_user(username = username)
            print "- User %s already exists." % username
        except UnknownObjectException, uoe:
            print "- Creating user %s" % username
            if not dry_run:
                user = ctx.obj["client"].create_user(username = username,
                                                     first_name = first_name,
                                                     last_name = last_name,
                                                     email = email)
        
        if cur_user_type == "student":
            try:
                student = course.get_student(username)
                if not student.dropped:
                    print "- User %s is already a student in %s" % (username, course_id)
                else:
                    if not dry_run:
                        student.dropped = False
                    print "- Student had previously been marked as dropped, has been un-dropped"
            except UnknownObjectException, uoe:
                print "- Adding student %s to %s" % (username, course_id)
                if not dry_run:
                    course.add_student(user)
        elif cur_user_type == "instructor":
            try:
                instructor = course.get_instructor(username)
                print "- User %s is already an instructor in %s" % (username, course_id)
            except UnknownObjectException, uoe:
                print "- Adding instructor %s to %s" % (username, course_id)
                if not dry_run:
                    course.add_instructor(user)
        elif cur_user_type == "grader":
            try:
                grader = course.get_grader(username)
                print "- User %s is already a grader in %s" % (username, course_id)
            except UnknownObjectException, uoe:
                print "- Adding grader %s to %s" % (username, course_id)
                if not dry_run:
                    course.add_grader(user)
            
        print 
    
    if sync:
        existing_students = course.get_students()
        for existing_student in existing_students:
            if existing_student.username not in student_usernames:
                if not dry_run:
                    existing_student.dropped = True
                print "Dropped %s" % existing_student.username
        

@click.command(name="create-git-users")
@click.argument('course_id', type=str)
@click.option('--staging', is_flag=True)
@click.option('--dry-run', is_flag=True)
@click.option('--all-types', is_flag=True)
@click.option('--only-type', type=click.Choice(VALID_USER_TYPES), required = False)
@require_config
@click.pass_context
def admin_course_create_git_users(ctx, course_id, staging, dry_run, all_types, only_type):   
    course = get_course_or_exit(ctx, course_id)    
                
    if all_types and only_type is not None:
        print "You cannot specify both --all-types and --only-type"
        ctx.exit(CHISUBMIT_FAIL)
        
    if staging and only_type == "student":
        print "You cannot create student accounts on the staging server."
        ctx.exit(CHISUBMIT_FAIL)
          
    conn = create_connection(course, ctx.obj['config'], staging)          
          
    users = []
    
    if not staging and (all_types or only_type == "student"):
        users += course.get_students()

    if all_types or only_type == "grader":
        users += course.get_graders()

    if all_types or only_type == "instructor":
        users += course.get_instructors()

    for user in users:
        if conn.exists_user(course, user):
            print "[SKIP] User '%s' already exists" % user.username
        else:
            if dry_run:
                print "[OK] Created user %s" % user.username
            else:
                try:
                    conn.create_user(course, user)
                    print "[OK] Created user %s" % user.username
                except ChisubmitException, ce:
                    print "[ERROR] Couldn't create user '%s': %s" % (user.username, ce.message)
    
@click.command(name="create-individual-teams")
@click.argument('course_id', type=str)
@click.option('--dry-run', is_flag=True)
@click.option('--only', type=str)
@require_config
@click.pass_context
def admin_course_create_individual_teams(ctx, course_id, dry_run, only):      
    course = get_course_or_exit(ctx, course_id)    

    if only is not None:
        students = [get_student_or_exit(ctx, course, only)]
    else:
        students = course.get_students()
        
    
    for student in students:
        print "Processing %s (%s, %s)" % (student.username, student.user.last_name, student.user.first_name)        

        try:
            team = course.get_team(student.username)

            tms = team.get_team_members()
            if len(tms) == 0:
                # Incomplete team creation
                if not dry_run:
                    team.add_team_member(student.username, confirmed = True)
                print "- Added user %s to team %s." % (student.username, team.team_id)
            elif len(tms) == 1 and tms[0].username == student.username:
                print "- User %s already has an individual team." % student.username
            else:
                print "- ERROR: Team '%s' exists but it has these team members: %s" % (team.team_id, [tm.username for tm in tms])
            
        except UnknownObjectException, uoe:
            if not dry_run:
                team = course.create_team(student.username)
                team.add_team_member(student.username, confirmed = True)
            print "- Created individual team for user %s." % student.username
            
        print
        
        
            
    
@click.command(name="set-attribute")
@click.argument('course_id', type=str)
@click.argument('attr_name', type=str)
@click.argument('attr_value', type=str)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_set_attribute(ctx, course_id, attr_name, attr_value):
    course = get_course_or_exit(ctx, course_id)
    
    api_obj_set_attribute(ctx, course, attr_name, attr_value)
    



@click.command(name="setup-repo")
@click.argument('course_id', type=str)
@click.option('--staging', is_flag=True)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_setup_repo(ctx, course_id, staging):
    course = get_course_or_exit(ctx, course_id)   

    conn = create_connection(course, ctx.obj['config'], staging)
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)

    conn.init_course(course)

    return CHISUBMIT_SUCCESS

@click.command(name="unsetup-repo")
@click.argument('course_id', type=str)
@click.option('--staging', is_flag=True)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_unsetup_repo(ctx, course_id, staging):
    course = get_course_or_exit(ctx, course_id)

    conn = create_connection(course, ctx.obj['config'], staging)
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)

    conn.deinit_course(course)

    return CHISUBMIT_SUCCESS

@click.command(name="update-repo-access")
@click.argument('course_id', type=str)
@click.option('--staging', is_flag=True)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_update_repo_access(ctx, course_id, staging):
    course = get_course_or_exit(ctx, course_id)
    
    conn = create_connection(course, ctx.obj['config'], staging)
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)
    
    conn.update_instructors(course)
    conn.update_graders(course)


@click.command(name="create-repos")
@click.argument('course_id', type=str)
@click.option('--staging', is_flag=True)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_create_repos(ctx, course_id, staging):
    course = get_course_or_exit(ctx, course_id)

    teams = course.get_teams()

    if len(teams) == 0:
        print "Course %s has no teams. No repositories to create." % course_id
        ctx.exit(CHISUBMIT_FAIL)  

    max_len = max([len(t.team_id) for t in teams])

    conn = create_connection(course, ctx.obj['config'], staging)
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)
    
    v = ctx.obj["verbose"]
    already_has_repository = 0
    warning = 0
    created = 0
    for team in sorted(teams, key=operator.attrgetter("team_id")):
        if conn.exists_team_repository(course, team):
            already_has_repository += 1
            if v: print "%-*s  SKIPPING. Already has a repository." % (max_len, team.team_id)
            continue

        team_members = team.get_team_members()
        unconfirmed_students = [tm for tm in team_members if not tm.confirmed]
        
        if len(unconfirmed_students) > 0:
            usernames = [tm.student.username for tm in unconfirmed_students]
            warning += 1
            print "%-*s  WARNING. Team has unconfirmed students: %s" % (max_len, team.team_id, ",".join(usernames))
            continue


        if not staging:        
            missing = []
            for tm in team_members:
                if course.git_usernames == "custom":
                    if tm.student.git_username is None:
                        missing.append(tm.student.username)
                    
            if len(missing) > 0:
                warning += 1
                print "%-20s WARNING. These students haven't set their git usernames: %s" % (team.id, ",".join(missing))
                continue
        
        try:
            conn.create_team_repository(course, team)
            created += 1
            print "%-20s CREATED" % team.team_id
        except Exception, e:
            print "%-20s Unexpected exception %s: %s" % (team.team_id, e.__class__.__name__, e.message)

    print
    print "Existing: %i" % already_has_repository
    print "Created : %i" % created
    print "Warnings: %i" % warning

    return CHISUBMIT_SUCCESS


@click.command(name="team-repo-create")
@click.argument('course_id', type=str)
@click.argument('team_id', type=str)
@click.option('--ignore-existing', is_flag=True)
@click.option('--public', is_flag=True)
@click.option('--staging', is_flag=True)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_team_repo_create(ctx, course_id, team_id, ignore_existing, public, staging):
    course = get_course_or_exit(ctx, course_id) 
    team = get_team_or_exit(ctx, course, team_id) 

    conn = create_connection(course, ctx.obj['config'], staging)
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)
    
    if conn.exists_team_repository(course, team) and not ignore_existing:
        print "Team %s already has a repository" % team_id
        ctx.exit(CHISUBMIT_FAIL)
    else:
        conn.create_team_repository(course, team, fail_if_exists = not ignore_existing, private = not public)

    return CHISUBMIT_SUCCESS


@click.command(name="team-repo-update")
@click.argument('course_id', type=str)
@click.argument('team_id', type=str)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_team_repo_update(ctx, course_id, team_id):
    course = get_course_or_exit(ctx, course_id) 
    team = get_team_or_exit(ctx, course, team_id) 

    #if team.github_repo is None:
    #    print "Team %s does not have a repository." % team.id
    #    return CHISUBMIT_FAIL

    conn = create_connection(course, ctx.obj['config'], staging = False)
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)

    conn.update_team_repository(course, team)
    return CHISUBMIT_SUCCESS


@click.command(name="team-repo-remove")
@click.argument('course_id', type=str)
@click.argument('team_id', type=str)
@click.option('--ignore-non-existing', is_flag=True)
@click.option('--staging', is_flag=True)
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def admin_course_team_repo_remove(ctx, course_id, team_id, ignore_non_existing, staging):
    course = get_course_or_exit(ctx, course_id) 
    team = get_team_or_exit(ctx, course, team_id) 

    conn = create_connection(course, ctx.obj['config'], staging)
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)

    if not conn.exists_team_repository(course, team) and not ignore_non_existing:
        print "WARNING: Cannot delete repository because it doesn't exist"
        ctx.exit(CHISUBMIT_FAIL)
    else:
        conn.delete_team_repository(course, team, fail_if_not_exists = not ignore_non_existing)

    return CHISUBMIT_SUCCESS


admin_course.add_command(shared_course_list)
admin_course.add_command(shared_course_set_user_attribute)

admin_course.add_command(admin_course_add)
admin_course.add_command(admin_course_remove)
admin_course.add_command(admin_course_show)
admin_course.add_command(admin_course_add_instructor)
admin_course.add_command(admin_course_add_grader)
admin_course.add_command(admin_course_add_student)
admin_course.add_command(admin_course_load_users)
admin_course.add_command(admin_course_create_git_users)
admin_course.add_command(admin_course_create_individual_teams)
admin_course.add_command(admin_course_set_attribute)
admin_course.add_command(admin_course_setup_repo)
admin_course.add_command(admin_course_unsetup_repo)
admin_course.add_command(admin_course_update_repo_access)
admin_course.add_command(admin_course_create_repos)
admin_course.add_command(admin_course_team_repo_create)
admin_course.add_command(admin_course_team_repo_update)
admin_course.add_command(admin_course_team_repo_remove)





#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

from chisubmit.client.types import ChisubmitAPIObject, Attribute, APIStringType,\
    APIIntegerType, APIDateTimeType, APIDecimalType, APIBooleanType,\
    APIObjectType, APIListType

class Assignment(ChisubmitAPIObject):

    _api_attributes = {"url": Attribute(name="url", 
                                       attrtype=APIStringType, 
                                       editable=False),  
                       
                       "rubric_url": Attribute(name="rubric_url", 
                                       attrtype=APIStringType, 
                                       editable=False),                         
                       
                       "assignment_id": Attribute(name="assignment_id", 
                                       attrtype=APIStringType, 
                                       editable=True),  
    
                       "name": Attribute(name="name", 
                                            attrtype=APIStringType, 
                                            editable=True),  
    
                       "deadline": Attribute(name="deadline", 
                                        attrtype=APIDateTimeType, 
                                        editable=True),  
    
                       "min_students": Attribute(name="min_students", 
                                                attrtype=APIIntegerType, 
                                                editable=True),  
     
                       "max_students": Attribute(name="max_students", 
                                               attrtype=APIIntegerType, 
                                               editable=True),
                       
                       "rubric_url": Attribute(name="rubric_url", 
                                                    attrtype=APIStringType, 
                                                    editable=False),                       
                      }
    
    def get_rubric_components(self):
        """
        :calls: GET /courses/:course/assignments/:assignment/rubric
        :rtype: List of :class:`chisubmit.client.assignment.RubricComponent`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            self.rubric_url
        )
        return [RubricComponent(self._api_client, headers, elem) for elem in data]    
    
    
    def create_rubric_component(self, description, points, order = None):
        """
        :calls: POST /courses/:course/assignments/:assignment/rubric/
        :param description: string
        :param points: float
        :param order: int
        :rtype: :class:`chisubmit.client.assignment.RubricComponent`
        """
        assert isinstance(description, (str, unicode)), description
        
        post_data = {"description": description,
                     "points": points}
        
        if order is not None:
            post_data["order"] = order
        
        headers, data = self._api_client._requester.request(
            "POST",
            self.rubric_url,
            data = post_data
        )
        return RubricComponent(self._api_client, headers, data)
    
    def register(self, students):
        """
        :calls: POST /courses/:course/assignments/:assignment/register
        :param students: list of string
        :rtype: :class:`chisubmit.client.assignment.RegistrationResponse`
        """
        assert isinstance(students, (list, tuple)), students
        
        post_data = { "students": students }
        
        headers, data = self._api_client._requester.request(
            "POST",
            self.url + "/register",
            data = post_data
        )
        return RegistrationResponse(self._api_client, headers, data)       
    
class RubricComponent(ChisubmitAPIObject):

    _api_attributes = {"url": Attribute(name="url", 
                                       attrtype=APIStringType, 
                                       editable=False),  

                       "id": Attribute(name="id", 
                                       attrtype=APIIntegerType, 
                                       editable=False),  
                       
                       "description": Attribute(name="name", 
                                                attrtype=APIStringType, 
                                                editable=True),  
    
                       "order": Attribute(name="order", 
                                          attrtype=APIIntegerType, 
                                          editable=True),  
    
                       "points": Attribute(name="points", 
                                           attrtype=APIDecimalType, 
                                           editable=True)
                                              
                      }        
    
class RegistrationResponse(ChisubmitAPIObject):
    
    _api_attributes = {

                       "new_team": Attribute(name="new_team", 
                                         attrtype=APIBooleanType, 
                                         editable=False),  
                       
                       "team": Attribute(name="team", 
                                         attrtype=APIObjectType("chisubmit.client.team.Team"), 
                                         editable=False),  

                       "team_members": Attribute(name="team_members", 
                                         attrtype=APIListType(APIObjectType("chisubmit.client.team.TeamMember")), 
                                         editable=False),

                       "registration": Attribute(name="registration", 
                                          attrtype=APIObjectType("chisubmit.client.team.Registration"), 
                                          editable=False),  
                                              
                      }
    
        


#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

import chisubmit.client.users
import chisubmit.client.assignment
import chisubmit.client.team
from chisubmit.client.types import ChisubmitAPIObject, Attribute, AttributeType,\
    APIStringType, APIIntegerType
from chisubmit.client.users import User
import datetime

class Course(ChisubmitAPIObject):
    
    _api_attributes = {"url": Attribute(name="url", 
                                   attrtype=APIStringType, 
                                   editable=False),    
    
                       "course_id": Attribute(name="course_id", 
                                       attrtype=APIStringType, 
                                       editable=True),    
        
                       "name": Attribute(name="name", 
                                         attrtype=APIStringType, 
                                         editable=True),    

                       "git_server_connstr": Attribute(name="git_server_connstr", 
                                                  attrtype=APIStringType, 
                                                  editable=True),    

                       "git_staging_connstr": Attribute(name="git_staging_connstr", 
                                                  attrtype=APIStringType, 
                                                  editable=True),    
        
                       "git_usernames": Attribute(name="git_usernames", 
                                                  attrtype=APIStringType, 
                                                  editable=True),    
        
                       "git_staging_usernames": Attribute(name="git_staging_usernames", 
                                                          attrtype=APIStringType, 
                                                          editable=True),    
        
                       "extension_policy": Attribute(name="extension_policy", 
                                                     attrtype=APIStringType, 
                                                     editable=True),    
        
                       "default_extensions": Attribute(name="default_extensions", 
                                                       attrtype=APIIntegerType, 
                                                       editable=True),
                       
                       "instructors_url": Attribute(name="instructors_url", 
                                                    attrtype=APIStringType, 
                                                    editable=False),
                       
                       "graders_url": Attribute(name="graders_url", 
                                                    attrtype=APIStringType, 
                                                    editable=False),       
                       
                       "students_url": Attribute(name="students_url", 
                                                    attrtype=APIStringType, 
                                                    editable=False),                                                                                

                       "assignments_url": Attribute(name="assignments_url", 
                                                    attrtype=APIStringType, 
                                                    editable=False),                                                                                
                       
                       "teams_url": Attribute(name="teams_url", 
                                                    attrtype=APIStringType, 
                                                    editable=False),                                                                                
                      }


    def get_instructors(self):
        """
        :calls: GET /courses/:course/instructors/
        :rtype: List of :class:`chisubmit.client.users.Instructor`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            "/courses/" + self.course_id + "/instructors/"
        )
        return [chisubmit.client.users.Instructor(self._api_client, headers, elem) for elem in data]

    def get_instructor(self, username):
        """
        :calls: GET /courses/:course/instructors/:instructor
        :rtype: :class:`chisubmit.client.users.Instructor`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            "/courses/" + self.course_id + "/instructors/" + username
        )
        return chisubmit.client.users.Instructor(self._api_client, headers, data)
    
    def add_instructor(self, user_or_username, git_username = None, git_staging_username = None):
        """
        :calls: POST /courses/:course/instructors/
        :rtype: :class:`chisubmit.client.users.Instructor`
        """
        
        assert isinstance(user_or_username, (str, unicode)) or isinstance(user_or_username, User) 
        
        if isinstance(user_or_username, (str, unicode)):
            username = user_or_username
        elif isinstance(user_or_username, User):
            username = user_or_username.username
        
        post_data = {"username": username}
        
        if git_username is not None:
            post_data["git_username"] = git_username
        if git_staging_username is not None:
            post_data["git_staging_username"] = git_staging_username
        
        headers, data = self._api_client._requester.request(
            "POST",
            "/courses/" + self.course_id + "/instructors/",
            data = post_data
        )
        return chisubmit.client.users.Instructor(self._api_client, headers, data)    
    
    def remove_instructor(self, user_or_username):
        """
        :calls: DELETE /courses/:course/instructors/:username
        :rtype: None
        """
        
        assert isinstance(user_or_username, (str, unicode)) or isinstance(user_or_username, User) or isinstance(user_or_username, chisubmit.client.users.Instructor)
        
        if isinstance(user_or_username, (str, unicode)):
            username = user_or_username
        elif isinstance(user_or_username, User):
            username = user_or_username.username
        elif isinstance(user_or_username, chisubmit.client.users.Instructor):
            username = user_or_username.user.username
        
        _ = self._api_client._requester.request(
            "DELETE",
            "/courses/" + self.course_id + "/instructors/" + username
        )
        return None
    
    def get_graders(self):
        """
        :calls: GET /courses/:course/graders/
        :rtype: List of :class:`chisubmit.client.users.Grader`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            "/courses/" + self.course_id + "/graders/"
        )
        return [chisubmit.client.users.Grader(self._api_client, headers, elem) for elem in data]    

    def get_grader(self, username):
        """
        :calls: GET /courses/:course/graders/:grader
        :rtype: :class:`chisubmit.client.users.Grader`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            "/courses/" + self.course_id + "/graders/" + username
        )
        return chisubmit.client.users.Grader(self._api_client, headers, data)    
    
    def add_grader(self, user_or_username, git_username = None, git_staging_username = None):
        """
        :calls: POST /courses/:course/graders/
        :rtype: :class:`chisubmit.client.users.Grader`
        """
        
        assert isinstance(user_or_username, (str, unicode)) or isinstance(user_or_username, User) 
        
        if isinstance(user_or_username, (str, unicode)):
            username = user_or_username
        elif isinstance(user_or_username, User):
            username = user_or_username.username
        
        post_data = {"username": username }
        
        if git_username is not None:
            post_data["git_username"] = git_username
        if git_staging_username is not None:
            post_data["git_staging_username"] = git_staging_username
        
        headers, data = self._api_client._requester.request(
            "POST",
            "/courses/" + self.course_id + "/graders/",
            data = post_data
        )
        return chisubmit.client.users.Grader(self._api_client, headers, data)        
    
    def remove_grader(self, user_or_username):
        """
        :calls: DELETE /courses/:course/graders/:username
        :rtype: None
        """
        
        assert isinstance(user_or_username, (str, unicode)) or isinstance(user_or_username, User) or isinstance(user_or_username, chisubmit.client.users.Grader)
        
        if isinstance(user_or_username, (str, unicode)):
            username = user_or_username
        elif isinstance(user_or_username, User):
            username = user_or_username.username
        elif isinstance(user_or_username, chisubmit.client.users.Grader):
            username = user_or_username.user.username
        
        _ = self._api_client._requester.request(
            "DELETE",
            "/courses/" + self.course_id + "/graders/" + username
        )
        return None
        
    
    def get_students(self):
        """
        :calls: GET /courses/:course/students/
        :rtype: List of :class:`chisubmit.client.users.Student`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            "/courses/" + self.course_id + "/students/"
        )
        return [chisubmit.client.users.Student(self._api_client, headers, elem) for elem in data]    
    
    def get_student(self, username):
        """
        :calls: GET /courses/:course/students/:grader
        :rtype: :class:`chisubmit.client.users.Student`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            "/courses/" + self.course_id + "/students/" + username
        )
        return chisubmit.client.users.Student(self._api_client, headers, data)        
    
    def add_student(self, user_or_username, git_username = None, extensions = None, dropped = None):
        """
        :calls: POST /courses/:course/students/
        :rtype: :class:`chisubmit.client.users.Student`
        """
        
        assert isinstance(user_or_username, (str, unicode)) or isinstance(user_or_username, User) 
        
        if isinstance(user_or_username, (str, unicode)):
            username = user_or_username
        elif isinstance(user_or_username, User):
            username = user_or_username.username
        
        post_data = {"username": username }
        
        if git_username is not None:
            post_data["git_username"] = git_username
        if extensions is not None:
            post_data["extensions"] = extensions
        if dropped is not None:
            post_data["dropped"] = dropped
                    
        headers, data = self._api_client._requester.request(
            "POST",
            "/courses/" + self.course_id + "/students/",
            data = post_data
        )
        return chisubmit.client.users.Student(self._api_client, headers, data)       
    
    def remove_student(self, user_or_username):
        """
        :calls: DELETE /courses/:course/students/:username
        :rtype: None
        """
        
        assert isinstance(user_or_username, (str, unicode)) or isinstance(user_or_username, User) or isinstance(user_or_username, chisubmit.client.users.Student)
        
        if isinstance(user_or_username, (str, unicode)):
            username = user_or_username
        elif isinstance(user_or_username, User):
            username = user_or_username.username
        elif isinstance(user_or_username, chisubmit.client.users.Student):
            username = user_or_username.user.username
        
        _ = self._api_client._requester.request(
            "DELETE",
            "/courses/" + self.course_id + "/students/" + username
        )
        return None      
    
    def get_assignments(self):
        """
        :calls: GET /courses/:course/assignments/
        :rtype: List of :class:`chisubmit.client.assignment.Assignment`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            "/courses/" + self.course_id + "/assignments/"
        )
        return [chisubmit.client.assignment.Assignment(self._api_client, headers, elem) for elem in data]    
    
    def get_assignment(self, assignment_id):
        """
        :calls: GET /courses/:course/assignments/:assignment/
        :rtype: List of :class:`chisubmit.client.assignment.Assignment`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            "/courses/" + self.course_id + "/assignments/" + assignment_id
        )
        return chisubmit.client.assignment.Assignment(self._api_client, headers, data)
    
    def create_assignment(self, assignment_id, name, deadline, min_students = None, max_students = None):
        """
        :calls: POST /courses/:course/assignments/
        :param assignment_id: string
        :param name: string
        :param deadline: string
        :param min_students: int
        :param max_students: int
        :rtype: :class:`chisubmit.client.assignment.Assignment`
        """
        assert isinstance(assignment_id, (str, unicode)), assignment_id
        assert isinstance(deadline, (str, unicode)) or isinstance(deadline, datetime.datetime), deadline
        
        # TODO: Convert/validate date
        if isinstance(deadline, (str, unicode)):
            # TODO: validate date
            deadline_str = deadline
        elif isinstance(deadline, datetime.datetime):
            deadline_str = deadline.isoformat(sep=" ")
        
        post_data = {"assignment_id": assignment_id,
                     "name": name,
                     "deadline": deadline_str}
        
        if min_students is not None:
            post_data["min_students"] = min_students
        if max_students is not None:
            post_data["max_students"] = max_students
        
        headers, data = self._api_client._requester.request(
            "POST",
            "/courses/" + self.course_id + "/assignments/",
            data = post_data
        )
        return chisubmit.client.assignment.Assignment(self._api_client, headers, data)    
    
    def get_teams(self):
        """
        :calls: GET /courses/:course/teams/
        :rtype: List of :class:`chisubmit.client.team.Team`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            self.teams_url
        )
        return [chisubmit.client.team.Team(self._api_client, headers, elem) for elem in data]        
    
    def get_team(self, team_id):
        """
        :calls: GET /courses/:course/teams/
        :rtype: :class:`chisubmit.client.team.Team`
        """
        
        assert isinstance(team_id, (str, unicode)), team_id
        
        headers, data = self._api_client._requester.request(
            "GET",
            self.teams_url + team_id
        )
        return chisubmit.client.team.Team(self._api_client, headers, data)    
    
    def create_team(self, team_id, extensions = None, active = None):
        """
        :calls: POST /courses/:course/teams/
        :param name: string
        :param extensions: int
        :param active: bool
        :rtype: :class:`chisubmit.client.team.Team`
        """
        assert isinstance(team_id, (str, unicode)), team_id

        post_data = { "team_id": team_id }
        
        if extensions is not None:
            post_data["extensions"] = extensions
        if active is not None:
            post_data["active"] = active
        
        headers, data = self._api_client._requester.request(
            "POST",
            self.teams_url,
            data = post_data
        )
        return chisubmit.client.team.Team(self._api_client, headers, data)        
    



import click

from chisubmit.repos.grading import GradingGitRepo
from chisubmit.common import CHISUBMIT_FAIL, CHISUBMIT_SUCCESS,\
    ChisubmitException, handle_unexpected_exception
from chisubmit.client.course import Course

from functools import update_wrapper

from dateutil.parser import parse
import operator
from chisubmit.client.exceptions import UnknownObjectException,\
    UnauthorizedException, BadRequestException, ChisubmitRequestException
from chisubmit.client.types import AttributeType
from requests.exceptions import ConnectionError, SSLError
from requests.packages.urllib3.exceptions import SSLError as SSLError_urllib3
from click.globals import get_current_context
from chisubmit.config import Config, ConfigDirectoryNotFoundException
from chisubmit.client import Chisubmit


def __load_config_and_client(require_local):
    ctx = get_current_context()
    
    try:
        ctx.obj["config"] = Config.get_config(ctx.obj["config_dir"], ctx.obj["work_dir"], ctx.obj["config_overrides"])
    except ConfigDirectoryNotFoundException:
        if not require_local:
            ctx.obj["config"] = Config.get_global_config(ctx.obj["config_overrides"])
        else:
            raise ChisubmitException("This command must be run in a directory configured to use chisubmit.")

    api_url = ctx.obj["config"].get_api_url()
    api_key = ctx.obj["config"].get_api_key()
    
    if api_url is None:
        raise ChisubmitException("Configuration value 'api-url' not found")

    if api_key is None:
        raise ChisubmitException("No chisubmit credentials were found!")

    ctx.obj["client"] = Chisubmit(api_key, base_url=api_url)    
    

def require_config(f):

    def new_func(*args, **kwargs):
        __load_config_and_client(require_local = False)
                
        return f(*args, **kwargs)

    return update_wrapper(new_func, f)


def require_local_config(f):
    
    def new_func(*args, **kwargs):
        __load_config_and_client(require_local = True)
                
        return f(*args, **kwargs)

    return update_wrapper(new_func, f)


def pass_course(f):
    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        course_id = ctx.obj["config"].get_course()

        if course_id is None:
            raise ChisubmitException("No course has been specified. Make sure you're in a directory configured "
                                     "to use chisubmit or, alternatively, use the '-c course=COURSE_ID' option.")
        else:
            try:
                course_obj = ctx.obj["client"].get_course(course_id = course_id)
            except UnknownObjectException:
                raise click.BadParameter("Course '%s' does not exist" % course_id)

            ctx.obj["course_obj"] = course_obj

            return ctx.invoke(f, ctx.obj["course_obj"], *args, **kwargs)

    return update_wrapper(new_func, f)


def catch_chisubmit_exceptions(f):
    
    def new_func(*args, **kwargs):
        ctx = get_current_context()
        try:
            return f(*args, **kwargs)
        except UnknownObjectException, uoe:
            print
            print "ERROR: There was an error processing this request"
            print
            print "URL: %s" % uoe.url
            print "HTTP method: %s" % uoe.method
            print "Error: Not found (404)"
            if ctx.obj["debug"]:
                print
                uoe.print_debug_info()
        except UnauthorizedException, ue:
            print
            print "ERROR: Your chisubmit credentials are invalid"
            print
            print "URL: %s" % ue.url
            print "HTTP method: %s" % ue.method
            print "Error: Unauthorized (401)"
            if ctx.obj["debug"]:
                print
                ue.print_debug_info()            
        except BadRequestException, bre:
            print
            print "ERROR: There was an error processing this request"
            print
            print "URL: %s" % bre.url
            print "HTTP method: %s" % bre.method
            print "Error(s):"
            bre.print_errors()
            if ctx.obj["debug"]:
                print
                bre.print_debug_info()
        except ChisubmitRequestException, cre:
            print "ERROR: chisubmit server returned an HTTP error"
            print
            print "URL: %s" % cre.url
            print "HTTP method: %s" % cre.method
            print "Status code: %i" % cre.status
            print "Message: %s" % cre.reason
            if ctx.obj["debug"]:
                print
                cre.print_debug_info()        
        except ConnectionError, ce:
            if isinstance(ce.message, (SSLError, SSLError_urllib3)):
                print "ERROR: SSL certificate error when connecting to server"                
            else:
                print "ERROR: Could not connect to server"
            print "URL: %s" % ce.request.url
            if ctx.obj["debug"]:
                print "Reason:", ce.message
        except ChisubmitException, ce:
            print "ERROR: %s" % ce.message
            if ctx.obj["debug"]:
                ce.print_exception()
        except click.UsageError:
            raise
        except Exception, e:
            handle_unexpected_exception()
            
        ctx.exit(CHISUBMIT_FAIL)
            
    return update_wrapper(new_func, f)


def get_course_or_exit(ctx, course_id):
    try:
        course = ctx.obj["client"].get_course(course_id = course_id)
        return course
    except UnknownObjectException:
        print "Course %s does not exist" % course_id
        ctx.exit(CHISUBMIT_FAIL)    

def get_user_or_exit(ctx, username):
    try:
        user = ctx.obj["client"].get_user(username = username)
        return user
    except UnknownObjectException:
        print "User %s does not exist" % username
        ctx.exit(CHISUBMIT_FAIL)    

def get_assignment_or_exit(ctx, course, assignment_id):
    try:
        return course.get_assignment(assignment_id = assignment_id)
    except UnknownObjectException:
        print "Assignment %s does not exist" % assignment_id
        ctx.exit(CHISUBMIT_FAIL)
        
def get_team_or_exit(ctx, course, team_id):
    try:
        return course.get_team(team_id = team_id)
    except UnknownObjectException:
        print "Team %s does not exist" % team_id
        ctx.exit(CHISUBMIT_FAIL)

def get_assignment_registration_or_exit(ctx, team, assignment_id):
    try:
        return team.get_assignment_registration(assignment_id = assignment_id)
    except UnknownObjectException:
        print "Team %s is not registered for assignment %s" % (team.team_id, assignment_id)
        ctx.exit(CHISUBMIT_FAIL)        
        
def get_instructor_or_exit(ctx, course, username):
    try:
        return course.get_instructor(username = username)
    except UnknownObjectException:
        print "Course %s does not have an instructor %s" % (course.course_id, username)
        ctx.exit(CHISUBMIT_FAIL)  
        
def get_grader_or_exit(ctx, course, username):
    try:
        return course.get_grader(username = username)
    except UnknownObjectException:
        print "Course %s does not have a grader %s" % (course.course_id, username)
        ctx.exit(CHISUBMIT_FAIL)  
        
def get_student_or_exit(ctx, course, username):
    try:
        return course.get_student(username = username)
    except UnknownObjectException:
        print "Course %s does not have a student %s" % (course.course_id, username)
        ctx.exit(CHISUBMIT_FAIL)                                


def api_obj_set_attribute(ctx, api_obj, attr_name, attr_value):
    valid_attrs = [attr for attr in api_obj._api_attributes.values() if attr.editable]
    valid_attrs_names = [attr.name for attr in valid_attrs]
    
    if attr_name not in valid_attrs_names:
        print "ERROR: '%s' is not a valid attribute." % attr_name
        print "Valid attributes are: %s" % (", ".join(valid_attrs_names))
        ctx.exit(CHISUBMIT_FAIL)
        
    attr = api_obj._api_attributes[attr_name]
    
    if attr.type.attrtype == AttributeType.STRING:
        v = attr_value
    elif attr.type.attrtype == AttributeType.INTEGER:
        v = int(attr_value)
    elif attr.type.attrtype == AttributeType.BOOLEAN:
        v = (attr_value in ("true", "True"))
    else:
        print "ERROR: Editing attribute '%s' from the command-line is not currently supported." % attr_name
        ctx.exit(CHISUBMIT_FAIL)
        
    api_obj.edit(**{attr_name: v}) 


class DateTimeParamType(click.ParamType):
    name = 'datetime'

    def convert(self, value, param, ctx):
        try:
            return parse(value)
        except ValueError:
            self.fail('"%s" is not a valid datetime string' % value, param, ctx)

DATETIME = DateTimeParamType()


def get_teams_registrations(course, assignment, only_ready_for_grading = False, grader = None, only = None):
    if only is not None:
        try:
            team = course.get_team(only)
            teams = [team]
        except UnknownObjectException:
            return {}
    else:
        teams = course.get_teams()

    # TODO: sideload the registrations
    rv = {}
    
    for team in teams:
        try:
            registration = team.get_assignment_registration(assignment.assignment_id)
            
            if (only_ready_for_grading and not registration.is_ready_for_grading()) or (grader is not None and registration.grader_username != grader.user.username):
                continue
            
            rv[team] = registration
        except UnknownObjectException:
            # Skip
            pass
        
    return rv


def create_grading_repos(config, course, assignment, teams_registrations):
    repos = []

    for team, registration in teams_registrations.items():
        repo = GradingGitRepo.get_grading_repo(config, course, team, registration)

        if repo is None:
            print ("Creating grading repo for %s... " % team.team_id),
            repo = GradingGitRepo.create_grading_repo(config, course, team, registration)
            repo.sync()

            repos.append(repo)

            print "done"
        else:
            print "Grading repo for %s already exists" % team.team_id

    return repos


def gradingrepo_push_grading_branch(config, course, team, registration, to_students=False, to_staging=False):
    repo = GradingGitRepo.get_grading_repo(config, course, team, registration)

    if repo is None:
        print "%s does not have a grading repository" % team.team_id
        return CHISUBMIT_FAIL
    
    if not repo.has_grading_branch():
        print "%s does not have a grading branch" % team.team_id
        return CHISUBMIT_FAIL

    if repo.is_dirty():
        print "Warning: %s grading repo has uncommitted changes." % team.team_id

    if to_students:
        repo.push_grading_branch_to_students()

    if to_staging:
        repo.push_grading_branch_to_staging()

    return CHISUBMIT_SUCCESS

def gradingrepo_pull_grading_branch(config, course, team, registration, from_students=False, from_staging=False):
    assert(not (from_students and from_staging))
    repo = GradingGitRepo.get_grading_repo(config, course, team, registration)

    if repo is None:
        print "%s does not have a grading repository" % team.team_id
        return CHISUBMIT_FAIL

    if repo.is_dirty():
        print "%s grading repo has uncommited changes. Cannot pull." % team.team_id
        return CHISUBMIT_FAIL

    if from_students:
        if not repo.has_grading_branch_staging():
            print "%s does not have a grading branch on students' repository" % team.team_id
        else:
            repo.pull_grading_branch_from_students()

    if from_staging:
        if not repo.has_grading_branch_staging():
            print "%s does not have a grading branch in staging" % team.team_id
        else:
            repo.pull_grading_branch_from_staging()

    return CHISUBMIT_SUCCESS




from defaults import *

SECRET_KEY = '0jd7t0(05@5d#-7$#w2#zagfruos0r!&t37%9m)_9^#ymg7dq+'

DEBUG = True

ALLOWED_HOSTS = []

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}



import click

import os.path
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL
from chisubmit.cli.common import create_grading_repos,\
    gradingrepo_push_grading_branch, gradingrepo_pull_grading_branch,\
    get_grader_or_exit, get_assignment_or_exit, get_teams_registrations,\
    catch_chisubmit_exceptions, require_local_config
from chisubmit.repos.grading import GradingGitRepo
from chisubmit.rubric import RubricFile, ChisubmitRubricException
from chisubmit.cli.common import pass_course
from chisubmit.cli.shared.course import shared_course_get_git_credentials


@click.group()
@click.pass_context
def grader(ctx):
    pass

@click.command(name="create-local-grading-repos")
@click.argument('grader_id', type=str)
@click.argument('assignment_id', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def grader_create_local_grading_repos(ctx, course, grader_id, assignment_id):
    grader = get_grader_or_exit(ctx, course, grader_id)
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    teams_registrations = get_teams_registrations(course, assignment, grader = grader)
    
    if len(teams_registrations) == 0:
        print "No teams found"
        ctx.exit(CHISUBMIT_FAIL)

    repos = create_grading_repos(ctx.obj['config'], course, assignment, teams_registrations)

    if not repos:
        print "There was some kind of problem creating the grading repos."
        ctx.exit(CHISUBMIT_FAIL)

    for repo in repos:
        repo.set_grader_author()

    for team, registration in teams_registrations.items():
        print ("Pulling grading branch for team %s... " % team.team_id),
        gradingrepo_pull_grading_branch(ctx.obj['config'], course, team, registration, from_staging = True)
        print "done"
        
    return CHISUBMIT_SUCCESS


@click.command(name="validate-rubrics")
@click.argument('grader_id', type=str)
@click.argument('assignment_id', type=str)
@click.option('--only', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def grader_validate_rubrics(ctx, course, grader_id, assignment_id, only):
    grader = get_grader_or_exit(ctx, course, grader_id)
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    teams_registrations = get_teams_registrations(course, assignment, grader = grader, only = only)
    
    for team, registration in teams_registrations.items():

        repo = GradingGitRepo.get_grading_repo(ctx.obj['config'], course, team, registration)
        if not repo:
            print "Repository for %s does not exist" % (team.team_id)
            ctx.exit(CHISUBMIT_FAIL)
    
        rubricfile = repo.repo_path + "/%s.rubric.txt" % assignment.assignment_id
    
        if not os.path.exists(rubricfile):
            print "Repository for %s does not exist have a rubric for assignment %s" % (team.team_id, assignment.assignment_id)
            ctx.exit(CHISUBMIT_FAIL)
    
        try:
            RubricFile.from_file(open(rubricfile), assignment)
            print "%s: Rubric OK." % team.team_id
        except ChisubmitRubricException, cre:
            print "%s: Rubric ERROR: %s" % (team.team_id, cre.message)

    return CHISUBMIT_SUCCESS


@click.command(name="push-grading-branches")
@click.argument('grader_id', type=str)
@click.argument('assignment_id', type=str)
@click.option('--only', type=str)
@require_local_config
@pass_course
@click.pass_context
def grader_push_grading_branches(ctx, course, grader_id, assignment_id, only):
    grader = get_grader_or_exit(ctx, course, grader_id)
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    teams_registrations = get_teams_registrations(course, assignment, grader = grader, only = only, only_ready_for_grading=True)
    
    if len(teams_registrations) == 0:
        print "No teams found"
        ctx.exit(CHISUBMIT_FAIL)

    for team, registration in teams_registrations.items():
        print "Pushing grading branch for team %s... " % team.team_id
        gradingrepo_push_grading_branch(ctx.obj['config'], course, team, registration, to_staging = True)

    return CHISUBMIT_SUCCESS

@click.command(name="pull-grading-branches")
@click.argument('grader_id', type=str)
@click.argument('assignment_id', type=str)
@click.option('--only', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def grader_pull_grading_branches(ctx, course, grader_id, assignment_id, only):
    grader = get_grader_or_exit(ctx, course, grader_id)
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    teams_registrations = get_teams_registrations(course, assignment, grader = grader, only = only, only_ready_for_grading=True)
    
    if len(teams_registrations) == 0:
        print "No teams found"
        ctx.exit(CHISUBMIT_FAIL)

    for team, registration in teams_registrations.items():
        print "Pulling grading branch for team %s... " % team.team_id
        gradingrepo_pull_grading_branch(ctx.obj['config'], course, team, registration, from_staging = True)

    return CHISUBMIT_SUCCESS

grader.add_command(grader_create_local_grading_repos)
grader.add_command(grader_validate_rubrics)
grader.add_command(grader_push_grading_branches)
grader.add_command(grader_pull_grading_branches)
grader.add_command(shared_course_get_git_credentials)




from django.http import Http404
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import status
from chisubmit.backend.api.models import Course, Student, Instructor, Grader,\
    Assignment, Team, RubricComponent, get_user_by_username, TeamMember,\
    Registration, Submission, Grade, CourseRoles
from chisubmit.backend.api.serializers import CourseSerializer,\
    StudentSerializer, InstructorSerializer, GraderSerializer,\
    AssignmentSerializer, TeamSerializer, UserSerializer,\
    RubricComponentSerializer, RegistrationRequestSerializer, RegistrationSerializer, TeamMemberSerializer,\
    RegistrationResponseSerializer, SubmissionSerializer,\
    SubmissionRequestSerializer, SubmissionResponseSerializer, GradeSerializer
from rest_framework.exceptions import PermissionDenied
from django.contrib.auth.models import User
from django.db import Error
from rest_framework.authentication import BasicAuthentication,\
    TokenAuthentication
from rest_framework.authtoken.models import Token
from chisubmit.common.utils import get_datetime_now_utc
from chisubmit.backend.api.helpers import get_course_person, get_assignment,\
    get_team, get_course, get_rubric_component, get_team_member,\
    get_registration, get_submission, get_grade

class CourseList(APIView):
    def get(self, request, format=None):
        courses = Course.objects.all()
        if not (request.user.is_staff or request.user.is_superuser):
            courses = [c for c in courses if c.has_user(request.user)]
        response_courses = []
        for course in courses:
            serializer = CourseSerializer(course, context={'request': request, 'course': course})
            response_courses.append(serializer.data)
        return Response(response_courses)

    def post(self, request, format=None):
        if not (request.user.is_staff or request.user.is_superuser):
            raise PermissionDenied
        
        serializer = CourseSerializer(data=request.data, context={'request': request})
        if serializer.is_valid():
            course_id = serializer.validated_data["course_id"]
            if Course.get_by_course_id(course_id) is not None:
                msg = "There is already a course with course_id = %s" % course_id
                return Response({"course_id": [msg]}, status=status.HTTP_400_BAD_REQUEST)
            try:
                serializer.save()
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

  
class CourseDetail(APIView):
            
    def get(self, request, course_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        serializer = CourseSerializer(course_obj, context=serializer_context)
        
        return Response(serializer.data)

    def patch(self, request, course_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        serializer = CourseSerializer(course_obj, data=request.data, partial=True, context=serializer_context)
        
        if serializer.is_valid():
            try:
                serializer.save()
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data)
        
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def delete(self, request, course_id, format=None):
        course_obj, roles = get_course(request, course_id)
        
        if not CourseRoles.ADMIN in roles:
            raise PermissionDenied
        
        course_obj.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)
                
 
class PersonList(APIView):
    def get(self, request, course_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles or CourseRoles.GRADER in roles):
            raise PermissionDenied
        
        persons = self.person_class.objects.filter(course = course_obj)
        
        serializer = self.person_serializer(persons, many=True, context=serializer_context)
        return Response(serializer.data)

    def post(self, request, course_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        if not CourseRoles.ADMIN in roles:
            raise PermissionDenied
        
        serializer = self.person_serializer(data=request.data, context=serializer_context)
        if serializer.is_valid():
            user = serializer.validated_data["user"]
            if self.person_class.objects.filter(course=course_obj, user=request.user).exists():
                return Response({"username": ["%s is already a %s in %s" % (user.username, self.person_str, course_obj.course_id)]}, status=status.HTTP_400_BAD_REQUEST)

            try:
                serializer.save(course=course_obj)
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)
    
class PersonDetail(APIView):
            
    def get_person(self, request, course_obj, roles, username):
        return get_course_person(course_obj, request.user, roles, self.person_class, username)
            
    def get(self, request, course_id, username, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}        
        
        person_obj = self.get_person(request, course_obj, roles, username)
        serializer_context["is_owner"] = (request.user.username == username)
        serializer = self.person_serializer(person_obj, context=serializer_context)
        return Response(serializer.data)

    def patch(self, request, course_id, username, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}        
        
        person_obj = self.get_person(request, course_obj, roles, username)
        serializer_context["is_owner"] = (request.user.username == username)

        serializer = self.person_serializer(person_obj, data=request.data, partial=True, context=serializer_context)        

        if serializer.is_valid():
            try:
                serializer.save()
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def delete(self, request, course_id, username, format=None):
        course_obj, roles = get_course(request, course_id)
        person_obj = self.get_person(request, course_obj, roles, username)
        
        if not (CourseRoles.ADMIN in roles):
            raise PermissionDenied
        
        person_obj.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)    
    

class InstructorList(PersonList):
    person_class = Instructor    
    person_serializer = InstructorSerializer
    person_str = "instructor"
    
class InstructorDetail(PersonDetail):
    person_class = Instructor    
    person_serializer = InstructorSerializer
    person_str = "instructor"    
    

class GraderList(PersonList):
    person_class = Grader    
    person_serializer = GraderSerializer
    person_str = "grader"
    
class GraderDetail(PersonDetail):
    person_class = Grader    
    person_serializer = GraderSerializer
    person_str = "grader"    
        
        
class StudentList(PersonList):
    person_class = Student    
    person_serializer = StudentSerializer
    person_str = "student"
    
class StudentDetail(PersonDetail):
    person_class = Student    
    person_serializer = StudentSerializer
    person_str = "student"    
    
    
class AssignmentList(APIView):
    def get(self, request, course_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}   
                
        assignments = Assignment.objects.filter(course = course_obj)
        serializer = AssignmentSerializer(assignments, many=True, context=serializer_context)
        return Response(serializer.data)

    def post(self, request, course_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}   

        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied

        serializer = AssignmentSerializer(data=request.data, context=serializer_context)
        if serializer.is_valid():
            assignment_id = serializer.validated_data["assignment_id"]
            if course_obj.get_assignment(assignment_id) is not None:
                msg = "There is already an assignment in course %s with assignment_id = %s" % (course_obj.course_id, assignment_id)
                return Response({"assignment_id": [msg]}, status=status.HTTP_400_BAD_REQUEST)   
            try:
                serializer.save(course=course_obj)
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)    
    

class AssignmentDetail(APIView):

    def get(self, request, course_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}   
        
        assignment_obj = get_assignment(course_obj, request.user, roles, assignment_id)
        serializer = AssignmentSerializer(assignment_obj, context=serializer_context)
        
        return Response(serializer.data)

    def patch(self, request, course_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}   
        
        assignment_obj = get_assignment(course_obj, request.user, roles, assignment_id)
        serializer = AssignmentSerializer(assignment_obj, data=request.data, partial=True, context=serializer_context)        

        if serializer.is_valid():
            try:
                serializer.save()
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def delete(self, request, course_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        
        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied        

        assignment_obj = get_assignment(course_obj, request.user, roles, assignment_id)
        assignment_obj.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)    
    

class RubricList(APIView):
   
    def get(self, request, course_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
                   
        assignment_obj = get_assignment(course_obj, request.user, roles, assignment_id)        
        rubric_components = assignment_obj.get_rubric_components()
        
        serializer = RubricComponentSerializer(rubric_components, many=True, context=serializer_context)
        return Response(serializer.data)

    def post(self, request, course_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
                   
        assignment_obj = get_assignment(course_obj, request.user, roles, assignment_id)        
        serializer = RubricComponentSerializer(data=request.data, context=serializer_context)
        if serializer.is_valid():
            description = serializer.validated_data["description"]
            rubric_description = assignment_obj.get_rubric_component_by_description(description)
            if rubric_description is not None:
                msg = "There is already a rubric component in assignment %s with description = '%s'" % (assignment_id, description)
                return Response({"description": [msg]}, status=status.HTTP_400_BAD_REQUEST)   
            try:
                serializer.save(assignment = assignment_obj)
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)    
    

class RubricDetail(APIView):

    def get(self, request, course_id, assignment_id, rubric_component_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
                   
        rubric_component_obj = get_rubric_component(course_obj, request.user, roles, assignment_id)   
        serializer = RubricComponentSerializer(rubric_component_obj, context=serializer_context)
        return Response(serializer.data)

    def patch(self, request, course_id, assignment_id, rubric_component_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
                   
        rubric_component_obj = get_rubric_component(course_obj, request.user, roles, assignment_id, rubric_component_id)   
        serializer = RubricComponentSerializer(rubric_component_obj, data=request.data, partial=True, context=serializer_context)

        if serializer.is_valid():
            try:
                serializer.save()
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def delete(self, request, course_id, assignment_id, rubric_component_id, format=None):
        course_obj, roles = get_course(request, course_id)
        rubric_component_obj = get_rubric_component(course_obj, request.user, roles, assignment_id, rubric_component_id)   

        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied
                
        rubric_component_obj.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)        


class Register(APIView):

    def post(self, request, course_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}

        assignment_obj = get_assignment(course_obj, request.user, roles, assignment_id)        
    
        # Adding dict() because apparently serializers.ListField won't stomach a QueryDict 
        serializer = RegistrationRequestSerializer(data=dict(request.data))
        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)              

        students_usernames = serializer.validated_data["students"]

        if len(roles) == 1 and CourseRoles.STUDENT in roles:
            is_student = True
            
            if request.user.username not in students_usernames:
                msg = "The list of students does not include you."
                return Response({"students": [msg]}, status=status.HTTP_400_BAD_REQUEST)
            
            user_student_obj = course_obj.get_student(request.user)
            student_objs = [user_student_obj]
            other_students = [s for s in students_usernames if s != request.user.username]
        else:
            is_student = False
            
            student_objs = []
            other_students = students_usernames

        students_errors = []
        for student in other_students:
            user_obj = get_user_by_username(student)
            
            if user_obj is not None:
                student_obj = course_obj.get_student(user_obj)
            else:
                student_obj = None
            
            if user_obj is None or student_obj is None:
                msg = "User '%s' is either not a valid user or not a student in course '%s'" % (student, course_obj.course_id)
                students_errors.append(msg)
            else:
                student_objs.append(student_obj)
                
        if len(students_errors) > 0:
            return Response({"students": students_errors}, status=status.HTTP_400_BAD_REQUEST)
                    
        if len(students_usernames) < assignment_obj.min_students or len(students_usernames) > assignment_obj.max_students:
            if assignment_obj.min_students == assignment_obj.max_students:
                if assignment_obj.min_students == 1:
                    msg = "You specified %i students, but this assignment only accepts individual registrations." % (len(students_usernames))                                    
                else: 
                    msg = "You specified %i students, but this assignment requires teams of %i students" % (len(students_usernames), assignment_obj.min_students)
            else:                
                msg = "You specified %i students, but this assignment requires teams with at least %i students and at most %i students" % (len(students_usernames), assignment_obj.min_students, assignment_obj.max_students)
            return Response({"students": [msg]}, status=status.HTTP_400_BAD_REQUEST)            
            
        create_team = False
        create_registration = False
        teams = course_obj.get_teams_with_students(student_objs)                
                
        if len(teams) > 0:
            # Teams that have the assignment, but are *not* a perfect match
            have_assignment = []
            students_have_assignment = set()
            perfect_match = None
            
            for t in teams:
                team_student_usernames = [u.user.username for u in t.students.all()]
                if len(team_student_usernames) == len(students_usernames):
                    match = True
                    for s in students_usernames:
                        if s not in team_student_usernames:
                            match = False
                            break
                        
                    if match:
                        if perfect_match is None:
                            perfect_match = t
                        else:
                            # There shouldn't be more than one perfect match
                            error_msg = "There is more than one team with the exact same students in it." \
                                        "Please notify your instructor."  
                            return Response({"fatal": [msg]}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)            
                    else:
                        if t.is_registered_for_assignment(assignment_obj):
                            have_assignment.append(t)
                            students_have_assignment.update([s for s in team_student_usernames if s in students_usernames])
                else:
                    if t.is_registered_for_assignment(assignment_obj):
                        have_assignment.append(t)
                        students_have_assignment.update([s for s in team_student_usernames if s in students_usernames])

            if len(have_assignment) > 0:
                error_msg = "'%s' is already registered for assignment '%s' in another team"
                error_msgs = [error_msg % (s, assignment_obj.assignment_id) for s in students_have_assignment]
                return Response({"students": error_msgs}, status=status.HTTP_400_BAD_REQUEST)                
                    
            if perfect_match is not None:
                team = perfect_match
                if perfect_match.is_registered_for_assignment(assignment_obj):
                    tm = perfect_match.teammember_set.get(student = user_student_obj)
                    tm.confirmed = True
                    tm.save()
                    
                    registration = perfect_match.get_registration(assignment_obj)
                else:
                    registration = Registration.objects.create(team = perfect_match,
                                                               assignment = assignment_obj)   
                    create_registration = True
            else:
                create_team = True                        
        else:
            create_team = True
            
        if create_team:
            team_id = "-".join(sorted(students_usernames))

            if course_obj.extension_policy == Course.EXT_PER_TEAM:
                default_extensions = course_obj.default_extensions
                extensions = default_extensions
            else:
                extensions = 0              
        
            team = Team.objects.create(course = course_obj,
                                       team_id = team_id,
                                       extensions = extensions)

            for student_obj in student_objs:
                if student_obj.user == request.user or not is_student:
                    confirmed = True
                else:
                    confirmed = False
                
                team_member = TeamMember.objects.create(student = student_obj,
                                                        team = team,
                                                        confirmed = confirmed)
                
            registration = Registration.objects.create(team = team,
                                                       assignment = assignment_obj)           
            
            create_registration = True     
                
        if create_team or create_registration:
            response_status = status.HTTP_201_CREATED
        else:
            response_status = status.HTTP_200_OK
            
        response_data = {"new_team": create_team,
                         "team": team,
                         "team_members": list(team.get_team_members()),
                         "registration": registration
                         }
            
        serializer = RegistrationResponseSerializer(response_data, context=serializer_context)            
        return Response(serializer.data, status=response_status)        
    

class TeamList(APIView):
    def get(self, request, course_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        if (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles or CourseRoles.GRADER in roles):
            teams = course_obj.get_teams()            
        elif len(roles) == 1 and CourseRoles.STUDENT in roles:
            student = course_obj.get_student(request.user)  
            teams = course_obj.get_teams_with_students([student])
        else:
            teams = []
        
        serializer = TeamSerializer(teams, many=True, context=serializer_context)
        return Response(serializer.data)

    def post(self, request, course_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied

        serializer = TeamSerializer(data=request.data, context=serializer_context)
        if serializer.is_valid():
            try:
                serializer.save(course=course_obj)
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)        
    
    
class TeamDetail(APIView):
            
    def get(self, request, course_id, team_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        team_obj = get_team(course_obj, request.user, roles, team_id)
        
        serializer = TeamSerializer(team_obj, context=serializer_context)
        return Response(serializer.data)

    def patch(self, request, course_id, team_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        team_obj = get_team(course_obj, request.user, roles, team_id)
        serializer = TeamSerializer(team_obj, data=request.data, partial=True, context=serializer_context)        

        if serializer.is_valid():
            serializer.save()
            return Response(serializer.data)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def delete(self, request, course_id, team_id, format=None):
        course_obj, roles = get_course(request, course_id)
        team_obj = get_team(course_obj, request.user, roles, team_id)

        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied        
        
        team_obj.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)        
    

class TeamMemberList(APIView):            
            
    def get(self, request, course_id, team_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        team_obj = get_team(course_obj, request.user, roles, team_id)
        
        serializer = TeamMemberSerializer(team_obj.teammember_set.all(), many=True, context=serializer_context)
        return Response(serializer.data)

    def post(self, request, course_id, team_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied        
        
        team_obj = get_team(course_obj, request.user, roles, team_id)
        
        serializer = TeamMemberSerializer(data=request.data, context=serializer_context)
        if serializer.is_valid():
            student = serializer.validated_data["student"]
            if TeamMember.objects.filter(team=team_obj, student=student).exists():
                return Response({"username": ["%s is already a member of team %s" % (student.user.username, team_obj.team_id)]}, status=status.HTTP_400_BAD_REQUEST)
            
            try:
                serializer.save(team = team_obj)
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)       


class TeamMemberDetail(APIView):
            
    def get(self, request, course_id, team_id, student_username, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
                
        teammember_obj = get_team_member(course_obj, request.user, roles, team_id, student_username)
        serializer = TeamMemberSerializer(teammember_obj, context=serializer_context)
        return Response(serializer.data)

    def patch(self, request, course_id, team_id, student_username, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
                
        teammember_obj = get_team_member(course_obj, request.user, roles, team_id, student_username)
        serializer = TeamMemberSerializer(teammember_obj, data=request.data, partial=True, context=serializer_context)        

        if serializer.is_valid():
            serializer.save()
            return Response(serializer.data)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def delete(self, request, course_id, team_id, student_username, format=None):
        course_obj, roles = get_course(request, course_id)              
        teammember_obj = get_team_member(course_obj, request.user, roles, team_id, student_username)

        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied
        
        teammember_obj.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)         
    
    
class RegistrationList(APIView):        
            
    def get(self, request, course_id, team_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        team_obj = get_team(course_obj, request.user, roles, team_id)
        
        serializer = RegistrationSerializer(team_obj.registration_set.all(), many=True, context=serializer_context)
        return Response(serializer.data)

    def post(self, request, course_id, team_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        team_obj = get_team(course_obj, request.user, roles, team_id)
        
        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied        
        
        serializer = RegistrationSerializer(data=request.data, context=serializer_context)
        if serializer.is_valid():
            try:
                serializer.save(team = team_obj)
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)       
    

class RegistrationDetail(APIView):

    def get(self, request, course_id, team_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        registration_obj = get_registration(course_obj, request.user, roles, team_id, assignment_id)
        
        serializer = RegistrationSerializer(registration_obj, context=serializer_context)
        return Response(serializer.data)

    def patch(self, request, course_id, team_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        registration_obj = get_registration(course_obj, request.user, roles, team_id, assignment_id)
        serializer = RegistrationSerializer(registration_obj, data=request.data, partial=True, context=serializer_context)        

        if serializer.is_valid():
            serializer.save()
            return Response(serializer.data)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def delete(self, request, course_id, team_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        registration_obj = get_registration(course_obj, request.user, roles, team_id, assignment_id)

        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied
        
        registration_obj.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)    
    

class Submit(APIView):

    def post(self, request, course_id, team_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        registration_obj = get_registration(course_obj, request.user, roles, team_id, assignment_id)
    
        serializer = SubmissionRequestSerializer(data=request.data)
        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)              

        now = get_datetime_now_utc()
                
        commit_sha = serializer.validated_data["commit_sha"]
        extensions_requested = serializer.validated_data["extensions"]
        ignore_deadline = serializer.validated_data["ignore_deadline"]
        
        dry_run = request.query_params.get("dry_run", "false") in ("true", "True")
        
        if ignore_deadline:
            if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
                msg = "Nice try! Only admins and instructors can ignore the deadline."
                return Response({"errors": [msg]}, status=status.HTTP_400_BAD_REQUEST)

        if not ignore_deadline and registration_obj.is_ready_for_grading():
            msg = "You cannot re-submit assignment %s." % (registration_obj.assignment.assignment_id)
            msg = " You made a submission before the deadline, and the deadline has passed."
            return Response({"errors": [msg]}, status=status.HTTP_400_BAD_REQUEST)
        
        submission = Submission(registration = registration_obj,
                                extensions_used = extensions_requested,
                                commit_sha = commit_sha,
                                submitted_at = now)
        
        valid, error_response, extensions = submission.validate()
        
        if not valid:
            return error_response
        else:
            if not dry_run:
                submission.save()
                registration_obj.final_submission = submission
                registration_obj.save()
                response_status = status.HTTP_201_CREATED
            else:
                response_status = status.HTTP_200_OK
            
            response_data = {"submission": submission,
                             "extensions_before": extensions["extensions_available_before"],
                             "extensions_after": extensions["extensions_available_after"]
                             }
                
            serializer = SubmissionResponseSerializer(response_data, context=serializer_context)            
            return Response(serializer.data, status=response_status)        
        

class SubmissionList(APIView):     
            
    def get(self, request, course_id, team_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        registration_obj = get_registration(course_obj, request.user, roles, team_id, assignment_id)
        
        serializer = SubmissionSerializer(registration_obj.submission_set.all(), many=True, context=serializer_context)
        return Response(serializer.data)

    def post(self, request, course_id, team_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        registration_obj = get_registration(course_obj, request.user, roles, team_id, assignment_id)

        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied
        
        serializer = SubmissionSerializer(data=request.data, context=serializer_context)
        if serializer.is_valid():
            try:
                serializer.save(registration = registration_obj)
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)       
    
class SubmissionDetail(APIView):
    
    def get(self, request, course_id, team_id, assignment_id, submission_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}

        submission_obj = get_submission(course_obj, request.user, roles, team_id, assignment_id, submission_id)
        serializer = SubmissionSerializer(submission_obj, context=serializer_context)
        return Response(serializer.data)

    def patch(self, request, course_id, team_id, assignment_id, submission_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}

        submission_obj = get_submission(course_obj, request.user, roles, team_id, assignment_id, submission_id)
        serializer = SubmissionSerializer(submission_obj, data=request.data, partial=True, context=serializer_context)        

        if serializer.is_valid():
            serializer.save()
            return Response(serializer.data)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def delete(self, request, course_id, team_id, assignment_id, submission_id, format=None):
        course_obj, roles = get_course(request, course_id)
        submission_obj = get_submission(course_obj, request.user, roles, team_id, assignment_id, submission_id)

        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied
        
        submission_obj.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)    

class GradeList(APIView):
            
    def get(self, request, course_id, team_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        registration_obj = get_registration(course_obj, request.user, roles, team_id, assignment_id)
        
        serializer = GradeSerializer(registration_obj.grade_set.all(), many=True, context=serializer_context)
        return Response(serializer.data)

    def post(self, request, course_id, team_id, assignment_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}
        
        registration_obj = get_registration(course_obj, request.user, roles, team_id, assignment_id)
        
        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied
        
        serializer = GradeSerializer(data=request.data, context=serializer_context)
        if serializer.is_valid():
            try:
                serializer.save(registration = registration_obj)
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)       
    
class GradeDetail(APIView): 
    
    def get(self, request, course_id, team_id, assignment_id, grade_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}

        grade_obj = get_grade(course_obj, request.user, roles, team_id, assignment_id, grade_id)
        serializer = GradeSerializer(grade_obj, context=serializer_context)
        return Response(serializer.data)

    def patch(self, request, course_id, team_id, assignment_id, grade_id, format=None):
        course_obj, roles = get_course(request, course_id)
        serializer_context = {'request': request, 'course': course_obj, 'roles': roles}

        grade_obj = get_grade(course_obj, request.user, roles, team_id, assignment_id, grade_id)
        serializer = GradeSerializer(grade_obj, data=request.data, partial=True, context=serializer_context)        
        
        if serializer.is_valid():
            serializer.save()
            return Response(serializer.data)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def delete(self, request, course_id, team_id, assignment_id, grade_id, format=None):
        course_obj, roles = get_course(request, course_id)
        grade_obj = get_grade(course_obj, request.user, roles, team_id, assignment_id, grade_id)

        if not (CourseRoles.ADMIN in roles or CourseRoles.INSTRUCTOR in roles):
            raise PermissionDenied
        
        grade_obj.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)    

    
class UserList(APIView):
    def get(self, request, format=None):
        if not (request.user.is_staff or request.user.is_superuser):
            raise PermissionDenied
        users = User.objects.all()
        serializer = UserSerializer(users, many=True, context={'request': request})
        return Response(serializer.data)

    def post(self, request, format=None):
        if not (request.user.is_staff or request.user.is_superuser):
            raise PermissionDenied
        serializer = UserSerializer(data=request.data, context={'request': request})
        if serializer.is_valid():
            username = serializer.validated_data["username"]
            try:
                user = User.objects.get(username=username)
                msg = "There is already a username with username = %s" % (username)
                return Response({"username": [msg]}, status=status.HTTP_400_BAD_REQUEST)                               
            except User.DoesNotExist:
                try:
                    serializer.save()
                except Error, e:
                    return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
                return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)    
    
class BaseUserDetail(APIView):
            
    def get_user(self, username):
        try:
            user = User.objects.get(username = username)
            return user
        except User.DoesNotExist:
            raise Http404  
            
    def _get(self, request, username, format=None):
        if username != request.user.username and not (request.user.is_staff or request.user.is_superuser):
            raise Http404        
        user = self.get_user(username)
        serializer = UserSerializer(user, context={'request': request})
        return Response(serializer.data)

    def _patch(self, request, username, format=None):
        if not (request.user.is_staff or request.user.is_superuser):
            if username == request.user.username:
                raise PermissionDenied
            else:
                raise Http404  
                  
        user = self.get_user(username)
        serializer = UserSerializer(user, data=request.data, partial=True, context={'request': request})        
        if serializer.is_valid():
            try:
                serializer.save()
            except Error, e:
                return Response({"database": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def _delete(self, request, username, format=None):
        if not (request.user.is_staff or request.user.is_superuser):
            if username == request.user.username:
                raise PermissionDenied
            else:
                raise Http404  
        user = self.get_user(username)
        user.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)    
    
class UserDetail(BaseUserDetail):    
    
    def get(self, request, username, format=None):
        return self._get(request, username, format)

    def patch(self, request, username, format=None):
        return self._patch(request, username, format)

    def delete(self, request, username, format=None):
        return self._delete(request, username, format)    
    
class AuthUserDetail(BaseUserDetail):    
    
    def get(self, request, format=None):
        return self._get(request, request.user.username, format)

    def patch(self, request, format=None):
        return self._patch(request, request.user.username, format)

    def delete(self, request, format=None):
        return self._delete(request, request.user.username, format)
    
class BaseUserToken(APIView):
    
    authentication_classes = (BasicAuthentication, TokenAuthentication)
    
    def _get(self, request, username, format=None):
        try:
            user = User.objects.get(username = username)
        except User.DoesNotExist:
            raise Http404          
        
        if not (request.user.is_staff or request.user.is_superuser or username == request.user.username):
            raise Http404
        
        reset = "true" in request.query_params.get("reset", [])
        
        if reset:
            try:
                token = Token.objects.get(user__username=username)
                old_token = token.key
                token.delete()
            except Token.DoesNotExist:
                old_token = None
        
            token = Token.objects.create(user=user)
            
            return Response({'old_token': old_token, 'token': token.key, 'new': True}, status=status.HTTP_201_CREATED)
        else:
            token, created = Token.objects.get_or_create(user=user)
        
            if created:
                return Response({'old_token': None, 'token': token.key, 'new': True}, status=status.HTTP_201_CREATED)
            else:
                return Response({'old_token': None, 'token': token.key, 'new': False}, status=status.HTTP_200_OK)
            
class UserToken(BaseUserToken):    
    
    def get(self, request, username, format=None):
        return self._get(request, username, format)

    
class AuthUserToken(BaseUserToken):    
    
    def get(self, request, format=None):
        return self._get(request, request.user.username, format)
        






#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

import click
from chisubmit.common import ChisubmitException, CHISUBMIT_FAIL,\
    handle_unexpected_exception, CHISUBMIT_SUCCESS
import sys
from pprint import pprint
from requests.exceptions import HTTPError, ConnectionError
from chisubmit.cli.admin import admin
from chisubmit.cli.instructor import instructor
from chisubmit.cli.student import student
from chisubmit.cli.grader import grader
import getpass
from docutils.utils.math.math2html import URL
from chisubmit.client.requester import BadRequestException
from chisubmit.client.exceptions import UnknownObjectException,\
    ChisubmitRequestException, UnauthorizedException
from functools import update_wrapper
from chisubmit.cli.common import catch_chisubmit_exceptions
import os
from chisubmit.repos.factory import RemoteRepositoryConnectionFactory
config = None

import chisubmit.common.log as log
from chisubmit.config import Config, ConfigDirectoryNotFoundException
from chisubmit import RELEASE
from chisubmit.client import Chisubmit


VERBOSE = False
DEBUG = False 

@click.group(name="chisubmit")
@click.option('--config', '-c', type=str, multiple=True)
@click.option('--config-dir', type=str, default=None)
@click.option('--work-dir', type=str, default=None)
@click.option('--verbose', '-v', is_flag=True)
@click.option('--debug', is_flag=True)
@click.version_option(version=RELEASE)
@catch_chisubmit_exceptions
@click.pass_context
def chisubmit_cmd(ctx, config, config_dir, work_dir, verbose, debug):
    global VERBOSE, DEBUG
    
    VERBOSE = verbose
    DEBUG = debug
    
    if config_dir is None and work_dir is not None:
        print "You cannot specify --work-dir without --config-dir"
        ctx.exit(CHISUBMIT_FAIL)

    if config_dir is not None and work_dir is None:
        print "You cannot specify --config-dir without --work-dir"
        ctx.exit(CHISUBMIT_FAIL)
    
    log.init_logging(verbose, debug)

    config_overrides = {}
    for c in config:
        if c.count("=") != 1 or c[0] == "=" or c[-1] == "=":
            raise click.BadParameter("Invalid configuration parameter: {}".format(c))
        else:
            k, v = c.split("=")
            config_overrides[k] = v

    ctx.obj = {}

    ctx.obj["config_overrides"] = config_overrides
    ctx.obj["config_dir"] = config_dir
    ctx.obj["work_dir"] = work_dir
    ctx.obj["verbose"] = verbose
    ctx.obj["debug"] = debug

    return CHISUBMIT_SUCCESS

chisubmit_cmd.add_command(admin)
chisubmit_cmd.add_command(instructor)
chisubmit_cmd.add_command(student)
chisubmit_cmd.add_command(grader)


@click.command(name="init")
@click.argument('course_id', required=False)
@click.option('--username', type=str)
@click.option('--password', type=str)
@click.option('--git-username', type=str)
@click.option('--git-password', type=str)
@click.option('--force', '-f', is_flag=True)
@catch_chisubmit_exceptions
@click.pass_context
def chisubmit_init(ctx, course_id, username, password, git_username, git_password, force):
                
    if ctx.obj["config_dir"] is None and ctx.obj["work_dir"] is None:
        try:
            config = Config.get_config()
            if not force:
                in_root = os.path.normpath(config.work_dir) == os.path.normpath(os.getcwd())
                
                if in_root:
                    print "The current directory is already configured to use chisubmit."
                else:
                    print "You are already inside a directory configured to use chisubmit."
                    print "Root directory: {}".format(config.work_dir)
                print
                if in_root:
                    print "If you're sure you want to reset the configuration of this directory,"
                    print "use the --force option"
                else:
                    print "If you're sure you want to create another chisubmit directory"
                    print "in the current directory, use the --force option"
                ctx.exit(CHISUBMIT_FAIL)
        except ConfigDirectoryNotFoundException:
            pass
    else:
        if not os.path.exists(ctx.obj["work_dir"]):
            print "The specified work directory does not exist: {}".format(ctx.obj["work_dir"])
            ctx.exit(CHISUBMIT_FAIL)      

        if os.path.exists(ctx.obj["config_dir"]) and not force:
            print "The specified configuration directory already exists: {}".format(ctx.obj["config_dir"])
            print "If you're sure you want to create a configuration directory there,"
            print "use the --force option."
            ctx.exit(CHISUBMIT_FAIL)      

    
    global_config = Config.get_global_config(config_overrides = ctx.obj["config_overrides"])
    
    api_url = global_config.get_api_url()
    api_key = global_config.get_api_key()
    
    if api_url is None:
        print "The 'api-url' configuration option is not set. I need this to connect to"
        print "the chisubmit server. If your instructor did not set this option"
        print "globally, you need to specify it like this:"
        print
        print "    chisubmit -c api-url=CHISUBMIT_API_URL init"
        print
        print "Where CHISUBMIT_API_URL should be replaced with the URL of the chisubmit server."
        print "Your instructor can provide you with the correct URL."
        ctx.exit(CHISUBMIT_FAIL)
        
    if api_key is None:
        guess_user = getpass.getuser()
        
        user_prompt = "Enter your chisubmit username [{}]: ".format(guess_user)
        password_prompt = "Enter your chisubmit password: "
        
        if username is None:
            username = raw_input(user_prompt)
            if len(username.strip()) == 0:
                username = guess_user
            
        if password is None:
            password = getpass.getpass(password_prompt)
        
        client = Chisubmit(username, password=password, base_url=api_url)
        
        try:
            api_key, _ = client.get_user_token()
        except UnauthorizedException:
            print "ERROR: Incorrect username/password"
            ctx.exit(CHISUBMIT_FAIL)
    
    client = Chisubmit(api_key, base_url=api_url)
    
    if course_id is not None:
        try:
            course = client.get_course(course_id = course_id)
        except UnknownObjectException:
            print "Cannot access course '{}'".format(course_id)
            print "This could mean the course does not exist, or you have not been added to this course."            
            ctx.exit(CHISUBMIT_FAIL)
    else:
        courses = client.get_courses()
        if len(courses) == 0:
            print "You do not have access to any courses on chisubmit."
            print "This could you have not been added to any course, or no courses have been created."
            ctx.exit(CHISUBMIT_FAIL)
            
        print
        print "You are a member of the following course(s)."
        print "Please select the one you want to use:"
        print
        n = 1
        for course in courses:
            print "[{}] {}: {}".format(n, course.course_id, course.name)
            n+=1
        print
        print "[X] Exit"
        print
        valid_options = [`x` for x in range(1, len(courses)+1)] + ['X', 'x']
        option = None
        while option not in valid_options:
            option = raw_input("Choose one: ")
            if option not in valid_options:
                print "'{}' is not a valid option!".format(option)
                print
        
        if option in ['X', 'x']:
            ctx.exit(CHISUBMIT_FAIL)
        else:
            course = courses[int(option)-1]
    
    connstr = course.git_server_connstr
    
    if connstr is None or connstr == "":
        print "Error: Course '{}' doesn't seem to be configured to use a Git server." % course.id
        ctx.exit(CHISUBMIT_FAIL)
        
    conn = RemoteRepositoryConnectionFactory.create_connection(connstr, staging = False)
    server_type = conn.get_server_type_name()

    # If there are global credentials defined, see if they work
    git_credentials = global_config.get_git_credentials(server_type)
    if git_credentials is not None:
        try:
            conn.connect(git_credentials)
        except ChisubmitException:
            git_credentials = None

    # If the existing credentials didn't work, get new ones
    if git_credentials is None:
        # Try the chisubmit username/password
        git_credentials, _ = conn.get_credentials(username, password)
        
        if git_credentials is None:
            user_prompt = "Enter your {} username: ".format(server_type)
            password_prompt = "Enter your {} password: ".format(server_type)
            
            if git_username is None:
                git_username = raw_input(user_prompt)
                
            if git_password is None:
                git_password = getpass.getpass(password_prompt)
            
            git_credentials, _ = conn.get_credentials(git_username, git_password)
    
            if git_credentials is None:
                print
                print "Unable to obtain {} credentials. Incorrect username/password.".format(server_type)

    config = Config.create_config_dir(ctx.obj["config_dir"], ctx.obj["work_dir"])
    
    # Always save the API URL locally
    config.set_api_url(api_url, where = Config.LOCAL)
    
    # Save the API key only if one was not found globally
    if config.get_api_key() is None:
        config.set_api_key(api_key, where = Config.LOCAL)
        
    # Always save the course locally
    config.set_course(course.course_id, where = Config.LOCAL)
    
    # Save the git credentials only if they're not the same as the global ones
    if config.get_git_credentials(server_type) != git_credentials:
        config.set_git_credentials(server_type, git_credentials, where = Config.LOCAL)
    
    print
    print "The following directory has been set up to use chisubmit"
    print "for course '{}' ({})".format(course.course_id, course.name)
    print
    print "   " + config.work_dir
    print
    print "Remember to run chisubmit only inside this directory"
    print "(or any of its subdirectories)"
    print
    
chisubmit_cmd.add_command(chisubmit_init)        





from django.core.urlresolvers import reverse
from rest_framework import status
from rest_framework.test import APITestCase
from django.contrib.auth.models import User


class AssignmentTests(APITestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_pa1']
    
    def test_get_assignments(self):
        user = User.objects.get(username='instructor1')
        self.client.force_authenticate(user=user)
        
        url = reverse('assignment-list', args=["cmsc40100"])
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        
    def test_get_assignment(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('assignment-detail', args=["cmsc40100","pa1"])
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        
    def test_create_assignment(self):
        user = User.objects.get(username='instructor1')
        self.client.force_authenticate(user=user)

        url = reverse('assignment-list', args=["cmsc40100"])
        
        post_data = {"assignment_id": "pa3",
                     "name": "Programming Assignment 3",
                     "deadline": "2042-02-04 20:00:00+00:00"}
        
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        


import click
from chisubmit.cli.common import pass_course, get_team_or_exit,\
    get_assignment_or_exit, catch_chisubmit_exceptions, require_local_config
from chisubmit.common import CHISUBMIT_FAIL, CHISUBMIT_SUCCESS,\
    ChisubmitException
import operator
from chisubmit.common.utils import convert_datetime_to_local

@click.command(name="list")
@click.option('--ids', is_flag=True)
@click.option('--assignment', type=str)
@click.option('--include-inactive', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def shared_team_list(ctx, course, ids, assignment, include_inactive):
    teams = course.get_teams()
    teams.sort(key=operator.attrgetter("team_id"))

    assignment_id = assignment
    if assignment_id is not None:
        assignment = get_assignment_or_exit(ctx, course, assignment_id)
    else:
        assignment = None

    for team in teams:
        registrations = team.get_assignment_registrations()
        
        if assignment is not None:
            if not assignment.assignment_id in [r.assignment.assignment_id for r in registrations]:
                continue
            
        if not (team.active or include_inactive):
            continue
        
        if ids:
            print team.team_id
        else:
            team_members = team.get_team_members()
            
            if len(team_members) == 0:
                students = "No students"
            else:
                students = "Students: " + ",".join([tm.student.user.username for tm in team_members])
            if len(registrations) == 0:
                assignments = "No assignments"
            else:
                assignments = "Assignments: " + ",".join([r.assignment.assignment_id for r in registrations])
            fields = [team.team_id, students, assignments]

            print "  ".join(fields)

    return CHISUBMIT_SUCCESS

@click.command(name="show")
@click.argument('team_id', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def shared_team_show(ctx, course, team_id):
    team = get_team_or_exit(ctx, course, team_id)
        
    print "Team name: %s" % team.team_id
    print
    if course.extension_policy == "per-team":
        print "Extensions available: %i" % team.extensions 
        print
    
    team_members = team.get_team_members()
    
    if len(team_members) == 0:
        print "No students in this team"
    else:
        print "STUDENTS"
        print "--------"
        for team_member in team_members:
            if team_member.confirmed:
                status = "CONFIRMED"
            else:
                status = "UNCONFIRMED"

            user = team_member.student.user

            print "%s: %s, %s  (%s)" % (user.username, user.last_name, user.first_name, status)

    print

    registrations = team.get_assignment_registrations()

    if len(registrations) == 0:
        print "This team is not registered for any assignments."
    else:
        print "ASSIGNMENTS"
        print "-----------"
        for r in registrations:
            assignment = r.assignment
            print "ID: %s" % assignment.assignment_id
            print "Name: %s" % assignment.name
            print "Deadline: %s" % convert_datetime_to_local(assignment.deadline).isoformat(" ")        
            #if ta.submitted_at is not None:
            #    print "Last submitted at: %s" % convert_datetime_to_local(ta.submitted_at).isoformat(" ")
            #    print "Commit SHA: %s" % ta.commit_sha
            #    print "Extensions used: %i" % ta.extensions_used
            #else:
            #    print "NOT SUBMITTED"    
            #print    



import click
from chisubmit.cli.common import pass_course, get_student_or_exit,\
    api_obj_set_attribute, catch_chisubmit_exceptions, require_local_config
from chisubmit.cli.shared.course import shared_course_list,\
    shared_course_get_git_credentials

@click.group(name="course")
@click.pass_context
def student_course(ctx):
    pass

@click.command(name="set-git-username")
@click.argument('git-username', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def student_course_set_git_username(ctx, course, git_username):
    user = ctx.obj["client"].get_user()
    
    student = get_student_or_exit(ctx, course, user.username)
    
    api_obj_set_attribute(ctx, student, "git_username", git_username)    
    

student_course.add_command(shared_course_list)
student_course.add_command(student_course_set_git_username)
student_course.add_command(shared_course_get_git_credentials)




from django.core.urlresolvers import reverse
from rest_framework import status
from rest_framework.test import APITestCase
from django.contrib.auth.models import User


class GradeTests(APITestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 
                         'course1_pa1', 'course1_pa1_registrations']    
            
    def test_create_grade(self):
        user = User.objects.get(username='instructor1')
        self.client.force_authenticate(user=user)

        url = reverse('grade-list', args=["cmsc40100","student1-student2","pa1"])
        
        post_data = {"rubric_component_id": 1,
                     "points": 30}
        
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        


from chisubmit.tests.integration.clientlibs import ChisubmitClientLibsTestCase
from chisubmit.tests.common import COURSE1_ASSIGNMENTS, COURSE2_ASSIGNMENTS,\
    COURSE1_TEAMS, COURSE1_TEAM_MEMBERS
from chisubmit.client.exceptions import BadRequestException
from chisubmit.backend.api.models import Course, Assignment, TeamMember,\
    Registration, Submission, Grade

class TeamTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams']
    
    def test_get_teams(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        teams = course.get_teams()
        
        self.assertEquals(len(teams), len(COURSE1_TEAMS))
        self.assertItemsEqual([t.team_id for t in teams], COURSE1_TEAMS)
        
    def test_get_team(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        
        for team_name in COURSE1_TEAMS:
            team = course.get_team(team_name)
            
            self.assertEquals(team.team_id, team_name)
            
    def test_create_team(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        team = course.create_team(team_id = "student2-student3",
                                  extensions = 2,
                                  active = False)
        self.assertEquals(team.team_id, "student2-student3")
        self.assertEquals(team.extensions, 2)
        self.assertEquals(team.active, False)
        
        course_obj = Course.get_by_course_id("cmsc40100")
        self.assertIsNotNone(course_obj)
        
        team_obj = course_obj.get_team("student2-student3")
        self.assertIsNotNone(team_obj, "Team was not added to database")
            
        self.assertEquals(team_obj.team_id, "student2-student3")                  
        self.assertEquals(team_obj.extensions, 2)                  
        self.assertEquals(team_obj.active, False)                  
                    
class TeamMemberTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams']
    
    def test_get_team_members(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        
        for t in COURSE1_TEAMS:        
            team = course.get_team(t)
            team_members = team.get_team_members()
            
            self.assertEquals(len(team_members), len(COURSE1_TEAM_MEMBERS[t]))
            self.assertItemsEqual([tm.username for tm in team_members], COURSE1_TEAM_MEMBERS[t])
            self.assertItemsEqual([tm.student.user.username for tm in team_members], COURSE1_TEAM_MEMBERS[t])
        
    def test_get_team_member(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        
        for t in COURSE1_TEAMS:        
            team = course.get_team(t)
            
            for tm in COURSE1_TEAM_MEMBERS[t]:
                team_member = team.get_team_member(tm)
                
                self.assertEqual(team_member.username, tm)
                self.assertEqual(team_member.student.user.username, tm)
                self.assertEqual(team_member.confirmed, True)
                
    
    def test_create_team_with_team_members(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        team = course.create_team(team_id = "student2-student3",
                                  extensions = 2,
                                  active = False)
        self.assertEquals(team.team_id, "student2-student3")
        self.assertEquals(team.extensions, 2)
        self.assertEquals(team.active, False)
        
        course_obj = Course.get_by_course_id("cmsc40100")
        self.assertIsNotNone(course_obj)
        
        team_obj = course_obj.get_team("student2-student3")
        self.assertIsNotNone(team_obj, "Team was not added to database")
            
        self.assertEquals(team_obj.team_id, "student2-student3")                  
        self.assertEquals(team_obj.extensions, 2)                  
        self.assertEquals(team_obj.active, False)
              
        team_member = team.add_team_member("student2", confirmed = False)
        self.assertEqual(team_member.username, "student2")
        self.assertEqual(team_member.student.user.username, "student2")
        self.assertEqual(team_member.confirmed, False)
                
        team_member_objs = TeamMember.objects.filter(team = team_obj, student__user__username = "student2")
        self.assertEquals(len(team_member_objs), 1)
        team_member_obj = team_member_objs[0]
        self.assertEqual(team_member_obj.student.user.username, "student2")
        self.assertEqual(team_member_obj.confirmed, False)
        
        team.add_team_member("student3", confirmed = True)
        team_member_objs = TeamMember.objects.filter(team = team_obj, student__user__username = "student3")
        self.assertEquals(len(team_member_objs), 1)
        team_member_obj = team_member_objs[0]
        self.assertEqual(team_member_obj.student.user.username, "student3")
        self.assertEqual(team_member_obj.confirmed, True)

        team_members = team.get_team_members()
        
        self.assertEquals(len(team_members), 2)
        self.assertItemsEqual([tm.username for tm in team_members], ["student2","student3"])
        self.assertItemsEqual([tm.student.user.username for tm in team_members], ["student2","student3"])        
        
class RegistrationTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 
                         'course1_pa1', 'course1_pa1_registrations',
                         'course1_pa2']    
    
    def test_get_registrations(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        
        for t in COURSE1_TEAMS:        
            team = course.get_team(t)
            registrations = team.get_assignment_registrations()
            
            self.assertEquals(len(registrations), 1)
            self.assertItemsEqual([r.assignment_id for r in registrations], ["pa1"])
            self.assertItemsEqual([r.assignment.assignment_id for r in registrations], ["pa1"])
        
    def test_get_registration(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        
        for t in COURSE1_TEAMS:        
            team = course.get_team(t)
            registration = team.get_assignment_registration("pa1")

            self.assertEqual(registration.assignment_id, "pa1")
            self.assertEqual(registration.assignment.assignment_id, "pa1")
            self.assertEqual(registration.grader_username, None)
            self.assertEqual(registration.grader, None)
            
    def test_create_registration(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        
        for t in COURSE1_TEAMS:        
            team = course.get_team(t)
            registration = team.add_assignment_registration("pa2")

            self.assertEqual(registration.assignment_id, "pa2")
            self.assertEqual(registration.assignment.assignment_id, "pa2")
            self.assertEqual(registration.grader_username, None)
            self.assertEqual(registration.grader, None)   
            
            registration_objs = Registration.objects.filter(team__team_id = t, assignment__assignment_id = "pa2")
            self.assertEquals(len(registration_objs), 1)
            registration_obj = registration_objs[0]
            self.assertEqual(registration_obj.team.team_id, t)
            self.assertEqual(registration_obj.assignment.assignment_id, "pa2")
            self.assertIsNone(registration_obj.grader)            
            
            registrations = team.get_assignment_registrations()
            
            self.assertEquals(len(registrations), 2)
            self.assertItemsEqual([r.assignment_id for r in registrations], ["pa1", "pa2"])
            self.assertItemsEqual([r.assignment.assignment_id for r in registrations], ["pa1", "pa2"])
            
class SubmissionTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 
                         'course1_pa1', 'course1_pa1_registrations_with_submissions',
                         'course1_pa2']
            
    def test_get_submissions(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        
        team = course.get_team("student1-student2")
        registration = team.get_assignment_registration("pa1")
        submissions = registration.get_submissions()
            
        self.assertEquals(len(submissions), 2)
        self.assertIn(registration.final_submission_id, [s.id for s in submissions])

        team = course.get_team("student3-student4")
        registration = team.get_assignment_registration("pa1")
        submissions = registration.get_submissions()
            
        self.assertEquals(len(submissions), 1)
        self.assertEquals(submissions[0].id, registration.final_submission_id)
        
    def test_get_submission(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        
        for t in COURSE1_TEAMS:        
            team = course.get_team(t)
            registration = team.get_assignment_registration("pa1")
            final_submission_id = registration.final_submission_id
            submission = registration.get_submission(final_submission_id)

            self.assertEqual(submission.id, final_submission_id)
            
    def test_create_submission(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        
        for t in COURSE1_TEAMS:        
            team = course.get_team(t)
            registration = team.get_assignment_registration("pa1")
            submissions = registration.get_submissions()
            n_submissions = len(submissions)
            
            sha = "COMMITSHA" + team.team_id
            submission = registration.add_submission(sha, 42)

            self.assertEqual(submission.extensions_used, 42)
            self.assertEqual(submission.commit_sha, sha)
            
            submission_objs = Submission.objects.filter(pk = submission.id)
            self.assertEquals(len(submission_objs), 1)
            submission_obj = submission_objs[0]
            self.assertEqual(submission_obj.extensions_used, 42)
            self.assertEqual(submission_obj.commit_sha, sha)
            
            submissions = registration.get_submissions()
            self.assertEquals(len(submissions), n_submissions + 1)
            
            
class GradeTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 
                         'course1_pa1', 'course1_pa1_registrations',
                         'course1_pa2']    
            
    def test_create_grade(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")

        assignment = course.get_assignment("pa1")
        rubric_components = assignment.get_rubric_components()
        
        points = 10
        
        for t in COURSE1_TEAMS:        
            team = course.get_team(t)
            registration = team.get_assignment_registration("pa1")
            
            for rc in rubric_components:
                grade = registration.add_grade(rc, points = points)

                self.assertEqual(grade.points, points)
                self.assertEqual(grade.rubric_component.description, rc.description)
                registration_obj = Registration.objects.get(team__team_id = team.team_id, assignment__assignment_id = "pa1")
                grade_objs = Grade.objects.filter(registration = registration_obj, rubric_component_id = rc.id)
                self.assertEquals(len(grade_objs), 1)
                grade_obj = grade_objs[0]
                self.assertEqual(grade_obj.points, points)
                
                points += 5
                
            
class ExistingGradeTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 
                     'course1_pa1', 'course1_pa1_registrations', 'course1_pa1_grades',
                     'course1_pa2']
            
    def test_get_grades(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        
        assignment = course.get_assignment("pa1")
        rubric_components = assignment.get_rubric_components()
        rubric_components_descriptions = [rc.description for rc in rubric_components]
        
        for t in COURSE1_TEAMS:        
            team = course.get_team(t)
            registration = team.get_assignment_registration("pa1")
            grades = registration.get_grades()
            
            self.assertEquals(len(grades), 2)
            self.assertItemsEqual([g.rubric_component.description for g in grades], rubric_components_descriptions)
        


from chisubmit.tests.integration.clientlibs import ChisubmitClientLibsTestCase
from chisubmit.client.exceptions import BadRequestException

class RegistrationTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 
                         'course1_pa1', 'course1_pa1_registrations',
                         'course1_pa2']
    
    def test_register_existing_team(self):
        c = self.get_api_client("student1token")
        
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa2")

        students = ["student1", "student2"]

        r = assignment.register(students = students)
                                
        self.assertEquals(r.new_team, False)          
        self.assertEquals(r.team.team_id, "student1-student2")
        self.assertItemsEqual([tm.username for tm in r.team_members], students)
        self.assertEquals(r.registration.assignment_id, "pa2")
        self.assertEquals(r.registration.assignment.assignment_id, "pa2")
        
    def test_register_new_team_single_student(self):
        c = self.get_api_client("student1token")
        
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa2")

        students = ["student1", "student3"]

        r = assignment.register(students = students)
                                
        self.assertEquals(r.new_team, True)          
        self.assertEquals(r.team.team_id, "student1-student3")
        self.assertItemsEqual([tm.username for tm in r.team_members], students)
        self.assertEquals(r.registration.assignment_id, "pa2")
        self.assertEquals(r.registration.assignment.assignment_id, "pa2")
        
    def test_register_new_team(self):
        students = ["student1", "student3"]

        c = self.get_api_client("student1token")
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa2")
        r = assignment.register(students = students)
                                
        self.assertEquals(r.new_team, True)          
        self.assertEquals(r.team.team_id, "student1-student3")
        self.assertItemsEqual([tm.username for tm in r.team_members], students)
        self.assertEquals(r.registration.assignment_id, "pa2")
        self.assertEquals(r.registration.assignment.assignment_id, "pa2")          

        c = self.get_api_client("student3token")
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa2")
        r = assignment.register(students = students)
                                
        self.assertEquals(r.new_team, False)          
        self.assertEquals(r.team.team_id, "student1-student3")
        self.assertItemsEqual([tm.username for tm in r.team_members], students)
        self.assertItemsEqual([tm.confirmed for tm in r.team_members], [True, True])
        self.assertEquals(r.registration.assignment_id, "pa2")
        self.assertEquals(r.registration.assignment.assignment_id, "pa2")
        
    def test_register_new_team_redundant(self):
        students = ["student1", "student3"]

        c = self.get_api_client("student1token")
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa2")
        r = assignment.register(students = students)
                                
        self.assertEquals(r.new_team, True)          
        self.assertEquals(r.team.team_id, "student1-student3")
        self.assertItemsEqual([tm.username for tm in r.team_members], students)
        self.assertEquals(r.registration.assignment_id, "pa2")
        self.assertEquals(r.registration.assignment.assignment_id, "pa2")          

        c = self.get_api_client("student3token")
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa2")
        r = assignment.register(students = students)
                                
        self.assertEquals(r.new_team, False)          
        self.assertEquals(r.team.team_id, "student1-student3")
        self.assertItemsEqual([tm.username for tm in r.team_members], students)
        self.assertItemsEqual([tm.confirmed for tm in r.team_members], [True, True])
        self.assertEquals(r.registration.assignment_id, "pa2")
        self.assertEquals(r.registration.assignment.assignment_id, "pa2")         
        
        c = self.get_api_client("student1token")
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa2")
        r = assignment.register(students = students)
                                
        self.assertEquals(r.new_team, False)          
        self.assertEquals(r.team.team_id, "student1-student3")
        self.assertItemsEqual([tm.username for tm in r.team_members], students)
        self.assertItemsEqual([tm.confirmed for tm in r.team_members], [True, True])
        self.assertEquals(r.registration.assignment_id, "pa2")
        self.assertEquals(r.registration.assignment.assignment_id, "pa2")                
        
    def test_register_non_student(self):
        c = self.get_api_client("instructor1token")
        
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa2")

        students = ["student1", "student2"]

        r = assignment.register(students = students)
                                
        self.assertEquals(r.new_team, False)          
        self.assertEquals(r.team.team_id, "student1-student2")
        self.assertItemsEqual([tm.username for tm in r.team_members], students)
        self.assertEquals(r.registration.assignment_id, "pa2")
        self.assertEquals(r.registration.assignment.assignment_id, "pa2") 
        
        
class RegistrationErrorTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 
                         'course1_pa1', 'course1_pa1_registrations',
                         'course1_pa2']
                
    def test_register_other_student(self):
        c = self.get_api_client("student1token")
        
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa2")

        students = ["student2", "student3"]

        with self.assertRaises(BadRequestException) as cm:
            r = assignment.register(students = students)
        
    def test_register_student_in_other_class(self):
        c = self.get_api_client("student1token")
        
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa2")

        students = ["student1", "student5"]

        with self.assertRaises(BadRequestException) as cm:
            r = assignment.register(students = students)
        
    def test_register_wrong_number_of_students(self):
        c = self.get_api_client("student1token")
        
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa2")

        students = ["student1", "student2", "student3"]

        with self.assertRaises(BadRequestException) as cm:
            r = assignment.register(students = students)        
            
    def test_register_multiple_groups(self):
        c = self.get_api_client("student1token")
        
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa1")

        students = ["student1", "student3"]

        with self.assertRaises(BadRequestException) as cm:
            r = assignment.register(students = students)               
        


#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

from chisubmit.client.types import ChisubmitAPIObject, Attribute, APIStringType, APIObjectType,\
    APIIntegerType, APIBooleanType

class User(ChisubmitAPIObject):
    _api_attributes = {"username": Attribute(name="username", 
                                            attrtype=APIStringType, 
                                            editable=False),    
    
                       "first_name": Attribute(name="first_name", 
                                              attrtype=APIStringType, 
                                              editable=False),    
    
                       "last_name": Attribute(name="name", 
                                             attrtype=APIStringType, 
                                             editable=False),    
    
                       "email": Attribute(name="email", 
                                         attrtype=APIStringType, 
                                         editable=False)
                      }    
    

class Instructor(ChisubmitAPIObject):

    _api_attributes = {"url": Attribute(name="url", 
                                       attrtype=APIStringType, 
                                       editable=False),  
    
                       "username": Attribute(name="username", 
                                            attrtype=APIStringType, 
                                            editable=False),  
    
                       "user": Attribute(name="user", 
                                        attrtype=APIObjectType(User), 
                                        editable=False),  
    
                       "git_username": Attribute(name="git_username", 
                                                attrtype=APIStringType, 
                                                editable=True),  
     
                       "git_staging_username": Attribute(name="git_staging_username", 
                                                        attrtype=APIStringType, 
                                                        editable=True)
                      }
    
class Grader(ChisubmitAPIObject):

    _api_attributes = {"url": Attribute(name="url", 
                                       attrtype=APIStringType, 
                                       editable=False),  
    
                       "username": Attribute(name="username", 
                                            attrtype=APIStringType, 
                                            editable=False),  
    
                       "user": Attribute(name="user", 
                                        attrtype=APIObjectType(User), 
                                        editable=False),  
    
                       "git_username": Attribute(name="git_username", 
                                                attrtype=APIStringType, 
                                                editable=True),  
     
                       "git_staging_username": Attribute(name="git_staging_username", 
                                                        attrtype=APIStringType, 
                                                        editable=True)
                      }
    
    # TODO
    def get_conflicts(self):
        return []
    
class Student(ChisubmitAPIObject):

    _api_attributes = {"url": Attribute(name="url", 
                                       attrtype=APIStringType, 
                                       editable=False),  
    
                       "username": Attribute(name="username", 
                                            attrtype=APIStringType, 
                                            editable=False),  
    
                       "user": Attribute(name="user", 
                                        attrtype=APIObjectType(User), 
                                        editable=False),  
    
                       "git_username": Attribute(name="git_username", 
                                                attrtype=APIStringType, 
                                                editable=True),  
     
                       "extensions": Attribute(name="extensions", 
                                               attrtype=APIIntegerType, 
                                               editable=True),
                         
                       "dropped": Attribute(name="dropped", 
                                            attrtype=APIBooleanType, 
                                            editable=True),  
                      }    


from chisubmit.tests.common import cli_test, ChisubmitCLITestClient, ChisubmitCLITestCase,\
    COURSE1_ID, COURSE1_NAME, COURSE2_ID, COURSE2_NAME
from chisubmit.backend.api.models import Course, Student, TeamMember
from django.contrib.auth.models import User

class CLIAdminCourse(ChisubmitCLITestCase):
            
    fixtures = ['admin_user']
            
    @cli_test
    def test_admin_course_add(self, runner):
        
        admin, _, _, _ = self.create_clients(runner, "admin")
        
        course_id = u"cmsc12300"
        course_name = u"Foobarmentals of Foobar"
        
        result = admin.run("admin course add", [course_id, course_name])
        self.assertEquals(result.exit_code, 0)
        
        try:
            course_obj = Course.objects.get(course_id=course_id)
        except Course.DoesNotExist:
            self.fail("Course was not added to database")  
            
        self.assertEquals(course_obj.course_id, course_id)                  
        self.assertEquals(course_obj.name, course_name) 
           
        
class CLIAdminCourseExisting(ChisubmitCLITestCase):
    
    fixtures = ['admin_user', 'course1', 'course2']    
        
    @cli_test
    def test_admin_course_list(self, runner):
        admin, _, _, _ = self.create_clients(runner, "admin")
        
        result = admin.run("admin course list")

        self.assertEquals(result.exit_code, 0)

        for val in [COURSE1_ID, COURSE1_NAME, COURSE2_ID, COURSE2_NAME]:            
            self.assertIn(val, result.output)

class CLIAdminCourseLoadUsers(ChisubmitCLITestCase):
    
    fixtures = ['admin_user', 'course1',  'course2']
             
    def gen_students(self, usernames):
        return {u: ("f"+u, "l"+u, u+"@example.org") for u in usernames}
    
    def gen_csv(self, students, filename):
        header = "username,first,last,email\n"
        rows = ["%s,%s,%s,%s" % (username, u[0], u[1], u[2]) for username, u in students.items()]
        rows.sort()
        rows = "\n".join(rows)
        
        with open(filename, "w") as f:
            f.write(header)
            f.write(rows)                 
                        
    @cli_test
    def test_admin_course_load_students(self, runner):
        admin, _, _, _ = self.create_clients(runner, "admin")
        
        students = self.gen_students(["student1", "student2", "student3", "student4"])
        csv_file = "students.csv"
        self.gen_csv(students, csv_file)
        
        result = admin.run("admin course load-users",
                           [COURSE1_ID, csv_file, "username", "first", "last", "email", "--user-type", "student"])
        self.assertEquals(result.exit_code, 0)

        user_objs = User.objects.all()
        student_objs = Student.objects.filter(course__course_id = COURSE1_ID)

        self.assertEquals(len(user_objs), len(students) + 1)
        self.assertEquals(len(student_objs), len(students))

        
    @cli_test
    def test_admin_course_load_students_twice(self, runner):
        admin, _, _, _ = self.create_clients(runner, "admin")
        
        students = self.gen_students(["student1", "student2", "student3", "student4"])
        csv_file = "students.csv"
        self.gen_csv(students, csv_file)
        
        result = admin.run("admin course load-users",
                           [COURSE1_ID, csv_file, "username", "first", "last", "email", "--user-type", "student"])        
        self.assertEquals(result.exit_code, 0)

        user_objs = User.objects.all()
        student_objs = Student.objects.filter(course__course_id = COURSE1_ID)

        self.assertEquals(len(user_objs), len(students) + 1)
        self.assertEquals(len(student_objs), len(students))

        result = admin.run("admin course load-users",
                           [COURSE1_ID, csv_file, "username", "first", "last", "email", "--user-type", "student"])        
        self.assertEquals(result.exit_code, 0)

        user_objs = User.objects.all()
        student_objs = Student.objects.filter(course__course_id = COURSE1_ID)

        self.assertEquals(len(user_objs), len(students) + 1)
        self.assertEquals(len(student_objs), len(students))
                
    @cli_test
    def test_admin_course_load_students_and_add(self, runner):
        admin, _, _, _ = self.create_clients(runner, "admin")
        
        students = self.gen_students(["student1", "student2", "student3", "student4"])
        csv_file = "students.csv"
        self.gen_csv(students, csv_file)
        
        result = admin.run("admin course load-users",
                           [COURSE1_ID, csv_file, "username", "first", "last", "email", "--user-type", "student"])        
        self.assertEquals(result.exit_code, 0)

        user_objs = User.objects.all()
        student_objs = Student.objects.filter(course__course_id = COURSE1_ID)

        self.assertEquals(len(user_objs), len(students) + 1)
        self.assertEquals(len(student_objs), len(students))

        students.update(self.gen_students(["student5"]))
        self.gen_csv(students, csv_file)

        result = admin.run("admin course load-users",
                           [COURSE1_ID, csv_file, "username", "first", "last", "email", "--user-type", "student"])        
        self.assertEquals(result.exit_code, 0)

        user_objs = User.objects.all()
        student_objs = Student.objects.filter(course__course_id = COURSE1_ID)

        self.assertEquals(len(user_objs), len(students) + 1)
        self.assertEquals(len(student_objs), len(students))
        
    @cli_test
    def test_admin_course_load_students_and_drop(self, runner):
        admin, _, _, _ = self.create_clients(runner, "admin")
        
        students = self.gen_students(["student1", "student2", "student3", "student4"])
        csv_file = "students.csv"
        self.gen_csv(students, csv_file)
        
        result = admin.run("admin course load-users",
                           [COURSE1_ID, csv_file, "username", "first", "last", "email", "--user-type", "student"])        
        self.assertEquals(result.exit_code, 0)

        user_objs = User.objects.all()
        student_objs = Student.objects.filter(course__course_id = COURSE1_ID)

        self.assertEquals(len(user_objs), len(students) + 1)
        self.assertEquals(len(student_objs), len(students))

        del students["student3"]
        self.gen_csv(students, csv_file)

        result = admin.run("admin course load-users",
                           [COURSE1_ID, csv_file, "username", "first", "last", "email", "--user-type", "student", "--sync"])        
        self.assertEquals(result.exit_code, 0)

        user_objs = User.objects.all()
        student_objs = Student.objects.filter(course__course_id = COURSE1_ID, dropped = False)

        self.assertEquals(len(user_objs), len(students) + 2)
        self.assertEquals(len(student_objs), len(students))        
        
        
    @cli_test
    def test_admin_course_load_students_two(self, runner):
        admin, _, _, _ = self.create_clients(runner, "admin")
        
        students1 = self.gen_students(["student1", "student2", "student3", "student4"])
        csv_file1 = "course1.csv"
        self.gen_csv(students1, csv_file1)
        
        result = admin.run("admin course load-users",
                           [COURSE1_ID, csv_file1, "username", "first", "last", "email", "--user-type", "student"])        
        self.assertEquals(result.exit_code, 0)

        user_objs = User.objects.all()
        student_objs = Student.objects.filter(course__course_id = COURSE1_ID)

        self.assertEquals(len(user_objs), len(students1) + 1)
        self.assertEquals(len(student_objs), len(students1))

        students2 = self.gen_students(["student3", "student4", "student5", "student6"])
        csv_file2 = "course2.csv"
        self.gen_csv(students2, csv_file2)
        
        result = admin.run("admin course load-users",
                           [COURSE2_ID, csv_file2, "username", "first", "last", "email", "--user-type", "student"])        
        self.assertEquals(result.exit_code, 0)

        user_objs = User.objects.all()
        student_objs = Student.objects.filter(course__course_id = COURSE2_ID)

        self.assertEquals(len(user_objs), 6 + 1)
        self.assertEquals(len(student_objs), len(students2))
             
          
class CLIAdminCourseCreateRepos(ChisubmitCLITestCase):
    
    fixtures = ['users', 'course1',  'course1_users', 'course1_teams']            
    
    def setup_git_testing(self, course_id):
        try:
            course_obj = Course.objects.get(course_id=course_id)
        except Course.DoesNotExist:
            self.fail("Course %s not in database" % course_id)
            
        course_obj.git_server_connstr = "server_type=Testing;local_path=./test-fs/server"
        course_obj.git_staging_connstr = "server_type=Testing;local_path=./test-fs/staging"
        course_obj.save()
    
    @cli_test
    def test_admin_course_create_repos(self, runner):
        admin, _, _, _ = self.create_clients(runner, "admin")
        
        self.setup_git_testing(COURSE1_ID)
        
        result = admin.run("admin course create-repos", [COURSE1_ID])          
        self.assertEquals(result.exit_code, 0)
        
    @cli_test
    def test_admin_course_create_repos_twice(self, runner):
        admin, _, _, _ = self.create_clients(runner, "admin")
        
        self.setup_git_testing(COURSE1_ID)
        
        result = admin.run("admin course create-repos", [COURSE1_ID])          
        self.assertEquals(result.exit_code, 0)        

        result = admin.run("admin course create-repos", [COURSE1_ID])          
        self.assertEquals(result.exit_code, 0)        

    @cli_test
    def test_admin_course_create_repos_incomplete_registration(self, runner):
        admin, _, _, _ = self.create_clients(runner, "admin")
        
        self.setup_git_testing(COURSE1_ID)
        
        tm = TeamMember.objects.get(team__team_id="student1-student2", student__user__username="student2")
        tm.confirmed = False
        tm.save()
        
        result = admin.run("admin course create-repos", [COURSE1_ID])          
        self.assertEquals(result.exit_code, 0)

        tm.confirmed = True
        tm.save()

        result = admin.run("admin course create-repos", [COURSE1_ID])          
        self.assertEquals(result.exit_code, 0)
    